import os
import oci
from oci.exceptions import CompositeOperationError, ServiceError


def _get_client_func(client, func):
    return getattr(client, func)


def _list(client, list_method_name, *args, **kwargs):
    if is_composite_client(client):
        func = _get_client_func(client.client, list_method_name)
    else:
        func = _get_client_func(client, list_method_name)
    items = perform_action(func, *args, **kwargs)
    if items:
        return items
    return []


def _perform_client_func_action(client, func_name, *args, **kwargs):
    if is_composite_client(client):
        func = _get_client_func(client.client, func_name)
    else:
        func = _get_client_func(client, func_name)
    return perform_action(func, *args, **kwargs)


def new_client(
    oci_client_class, composite_class=None, profile_name="DEFAULT", **kwargs
):
    # Try from environment, if not rely on config file
    file_location = oci.config.DEFAULT_LOCATION

    if "OCI_CONFIG_PATH" in os.environ:
        file_location = os.environ["OCI_CONFIG_PATH"]

    if "OCI_PROFILE_NAME" in os.environ:
        profile_name = os.environ["OCI_PROFILE_NAME"]

    client_config = None
    # Try to load from the config_file instead
    file_config = oci.config.from_file(
        file_location=file_location, profile_name=profile_name
    )

    # If an environment path for the key file is provided
    if "OCI_KEY_FILE" in os.environ:
        key_file = os.environ["OCI_KEY_FILE"]
        file_config.update(dict(key_file=key_file))

    oci.config.validate_config(file_config)
    client_config = file_config

    if not client_config:
        raise ValueError("The OCI client config must be loaded at this point")

    client = oci_client_class(client_config, **kwargs)
    if composite_class:
        return composite_class(client, **kwargs)
    return client


def is_composite_client(client):
    """ Composite clients have the 'client' attribute """
    return hasattr(client, "client")


def is_base_client(client):
    """Base clients have the 'base_client' attribute"""
    return hasattr(client, "base_client")


def list_entities(client, list_method, *args, **kwargs):
    return _list(client, list_method, *args, **kwargs)


def delete(client, delete_method, id, *args, wait_for_states=None, **kwargs):
    if not wait_for_states:
        wait_for_states = []

    action_kwargs = kwargs
    if is_composite_client(client):
        # Convert to the matching composite action
        pos_delete_method = "{}_and_wait_for_state".format(delete_method)
        if hasattr(client, pos_delete_method):
            delete_method = pos_delete_method
            action_kwargs["wait_for_states"] = wait_for_states
        else:
            # Fall back to the regular client since the
            # composite client dosen't have the method
            client = client.client

    func = _get_client_func(client, delete_method)
    return perform_action(func, id, **action_kwargs)


def create(client, create_method, *args, wait_for_states=None, **kwargs):
    if not wait_for_states:
        wait_for_states = []

    action_kwargs = kwargs
    if is_composite_client(client):
        # Convert to the matching composite action
        pos_create_method = "{}_and_wait_for_state".format(create_method)
        if hasattr(client, pos_create_method):
            create_method = pos_create_method
            action_kwargs["wait_for_states"] = wait_for_states
        else:
            # Fall back to the regular client since the
            # composite client dosen't have the method
            client = client.client

    func = _get_client_func(client, create_method)
    return perform_action(func, *args, **action_kwargs)


def update(client, update_method, id, wait_for_states=None, **kwargs):
    if not wait_for_states:
        wait_for_states = []

    action_kwargs = kwargs
    if is_composite_client(client):
        # Convert to the matching composite action
        pos_update_method = "{}_and_wait_for_state".format(update_method)
        if hasattr(client, pos_update_method):
            update_method = pos_update_method
            action_kwargs["wait_for_states"] = wait_for_states
        else:
            # Fall back to the regular client since the
            # composite client dosen't have the method
            client = client.client

    func = _get_client_func(client, update_method)
    return perform_action(func, id, **action_kwargs)


def get(client, get_method, id, **kwargs):
    return _perform_client_func_action(client, get_method, id, **kwargs)


def get_subnet_gateway_id(network_client, vcn_id, subnet_id, compartment_id):
    existing_subnets = network_client.list_subnets(compartment_id, vcn_id).data
    if not existing_subnets:
        return None
    for subnet in existing_subnets:
        if subnet.id != subnet_id:
            continue

        # Get the gateway of the first subnet
        subnet = existing_subnets[0]
        route_table = network_client.get_route_table(subnet.route_table_id).data
        # Find the Internet Gateway id in the first route for now
        route = route_table.route_rules[0]
    return route.network_entity_id


def prepare_route_rule(network_entity_id, **route_rule_kwargs):
    route_rule = oci.core.models.RouteRule(
        network_entity_id=network_entity_id, **route_rule_kwargs
    )
    return route_rule


def get_kubernetes_version(container_engine_client):
    """
    get_kubernetes_version

    Get the supported kubernetes versions from the service.  There are multiple
    versions supported but for the example we will just use the last one in the
    list.
    """

    if is_composite_client(container_engine_client):
        func = _get_client_func(container_engine_client.client, "get_cluster_options")
    else:
        func = _get_client_func(container_engine_client, "get_cluster_options")

    response = func(cluster_option_id="all")
    versions = response.data.kubernetes_versions
    if len(versions) > 0:
        kubernetes_version = versions[-1]
    else:
        raise RuntimeError("No supported Kubernetes versions found")

    return kubernetes_version


def stack_was_deleted(stack):
    for key_type, value_type in stack.items():
        if isinstance(value_type, (bool, str)):
            if not value_type:
                return False
        if isinstance(value_type, (list, tuple)):
            for result in value_type:
                if not result:
                    return False
    return True


def perform_action(action, *args, **kwargs):
    try:
        result = action(*args, **kwargs)
        if hasattr(result, "data"):
            return result.data
    # TODO, check for regular errors as well
    except (CompositeOperationError, ServiceError) as err:
        print("Failed to perform action: {}, err: {}".format(action, err))
        if hasattr(err, "cause"):
            print("Failed cause: {}".format(err.cause))

        return False
    return True



import os
import unittest
from corc.oci.instance import OCIInstanceOrchestrator


class TestInstanceOrchestrator(unittest.TestCase):
    def setUp(self):
        # Load compartment_id from the env
        if "OCI_COMPARTMENT_ID" not in os.environ:
            raise ValueError("Missing required environment variable OCI_COMPARTMENT_ID")

        if "OCI_PROFILE_NAME" in os.environ:
            profile_name = os.environ["OCI_PROFILE_NAME"]
        else:
            profile_name = "DEFAULT"

        oci_options = dict(
            compartment_id=os.environ["OCI_COMPARTMENT_ID"], profile_name=profile_name,
        )

        test_name = "Test_Instance_Orch"
        node_name = test_name + "_Node"
        vcn_name = test_name + "_Network"
        subnet_name = test_name + "_Subnet"

        # Add unique test postfix
        if "OCI_TEST_ID" in os.environ:
            node_name += os.environ["OCI_TEST_ID"]
            vcn_name += os.environ["OCI_TEST_ID"]
            subnet_name += os.environ["OCI_TEST_ID"]

        compute_options = dict(
            availability_domain="lfcb:EU-FRANKFURT-1-AD-1",
            shape="VM.Standard1.1",
            operating_system="CentOS",
            operating_system_version="7",
            display_name=node_name,
        )

        vcn_options = dict(display_name=vcn_name, dns_label="ku",)
        subnet_options = dict(display_name=subnet_name, dns_label="workers")

        self.options = dict(
            oci=oci_options,
            compute=compute_options,
            vcn=vcn_options,
            subnet=subnet_options,
        )

        OCIInstanceOrchestrator.validate_options(self.options)
        self.orchestrator = OCIInstanceOrchestrator(self.options)
        # Should not be ready at this point
        self.assertFalse(self.orchestrator.is_ready())

    def tearDown(self):
        self.orchestrator.tear_down()
        self.assertFalse(self.orchestrator.is_ready())
        self.orchestrator = None
        self.options = None

    def test_setup_instance(self):
        self.orchestrator.setup()
        self.assertTrue(self.orchestrator.is_ready())

    def test_teardown_instance(self):
        self.assertFalse(self.orchestrator.is_ready())
        self.orchestrator.tear_down()
        self.assertFalse(self.orchestrator.is_ready())


if __name__ == "__main__":
    unittest.main()



import boto3
import botocore
from botocore.configloader import load_config
from botocore.credentials import SharedCredentialProvider
import copy
import os


required_s3_fields = {
    "config_file": str,
    "credentials_file": str,
    "profile_name": str,
}

required_s3_values = {
    "config_file": True,
    "credentials_file": True,
    "profile_name": True,
}

required_s3_bucket_fields = {
    "bucket_name": str,
    "bucket_input_prefix": str,
    "bucket_output_prefix": str,
}

required_s3_bucket_values = {
    "bucket_name": True,
    "bucket_input_prefix": False,
    "bucket_output_prefix": False,
}


def upload_to_s3(s3_client, local_path, s3_path, bucket_name):
    if not os.path.exists(local_path):
        return False
    s3_client.upload_file(local_path, bucket_name, s3_path)
    return True


def upload_directory_to_s3(client, path, bucket_name, s3_prefix="output"):
    if not os.path.exists(path):
        return False
    for root, dirs, files in os.walk(path):
        for filename in files:
            local_path = os.path.join(root, filename)
            # generate s3 dir path
            # Skip the first /
            # HACK
            s3_path = local_path.split(path)[1][1:]
            # Upload
            if s3_prefix:
                s3_path = os.path.join(s3_prefix, s3_path)
            try:
                client.upload_file(local_path, bucket_name, s3_path)
            except Exception:
                return False
    return True


def bucket_exists(s3_client, bucket_name):
    try:
        s3_client.head_bucket(Bucket=bucket_name)
        return True
    except botocore.exceptions.ClientError:
        pass
    return False


def create_bucket(s3_client, bucket_name, **kwargs):
    bucket = s3_client.create_bucket(Bucket=bucket_name, **kwargs)
    if not bucket:
        return False
    return bucket


def delete_bucket(s3_client, bucket_name, **kwargs):
    return s3_client.delete_bucket(Bucket=bucket_name, **kwargs)


def delete_objects(s3_resource, bucket_name, s3_prefix="", max_chunk_size=1000):
    bucket = s3_resource.Bucket(bucket_name)
    objects_keys = [
        {"Key": str(obj.key)} for obj in bucket.objects.filter(Prefix=s3_prefix)
    ]

    if not objects_keys:
        return {}

    print(len(objects_keys))
    num_delete = len(objects_keys)

    if num_delete > max_chunk_size:
        delete_chunks = [
            objects_keys[i: i + max_chunk_size]
            for i in range(0, num_delete, max_chunk_size)
        ]
        for idx, chunk_keys in enumerate(delete_chunks):
            s3_resource.meta.client.delete_objects(
                Bucket=bucket_name, Delete={"Objects": chunk_keys}
            )

    return s3_resource.meta.client.delete_objects(
        Bucket=bucket_name, Delete={"Objects": objects_keys}
    )


# Accept parameters to
def expand_s3_bucket(s3_resource, bucket_name, target_dir=None, s3_prefix="input"):
    """ Expands the content of the specified bucket into the target_dir """
    bucket = s3_resource.Bucket(bucket_name)
    for obj in bucket.objects.filter(Prefix=s3_prefix):
        obj_path = copy.deepcopy(obj.key)
        if s3_prefix:
            obj_path = os.path.relpath(obj_path, s3_prefix)

        if target_dir:
            full_path = os.path.join(target_dir, obj_path)
        else:
            full_path = obj_path
        dir_path = os.path.dirname(full_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        bucket.download_file(obj.key, full_path)
    return True


def list_objects(s3_resource, bucket_name, all_objects=True, **kwargs):
    finished = False
    results = []
    last_object = None
    while not finished:
        response = None
        if last_object:
            response = s3_resource.meta.client.list_objects_v2(
                Bucket=bucket_name, StartAfter=last_object["Key"], **kwargs
            )
        else:
            response = s3_resource.meta.client.list_objects_v2(
                Bucket=bucket_name, **kwargs
            )

        if response and "Contents" in response:
            results.extend(response["Contents"])

        if not all_objects:
            finished = True
        else:
            if response and "Contents" in response and response["IsTruncated"]:
                last_object = response["Contents"][-1]
            else:
                finished = True
    return results


def stage_s3_resource(**kwargs):
    return boto3.resource("s3", **kwargs)


def load_s3_config(config_file, credentials_file, endpoint_url, profile_name="DEFAULT"):
    s3_config = load_config(config_file)
    if not s3_config:
        raise RuntimeError("Failed to load s3 config: {}".format(config_file))

    s3_creds_provider = SharedCredentialProvider(
        credentials_file, profile_name=profile_name,
    )
    s3_creds_config = s3_creds_provider.load()
    if not s3_creds_config:
        raise RuntimeError(
            "Failed to load s3 credentials config: {}".format(credentials_file)
        )

    (s3_access_key, s3_secret_key, s3_token,) = s3_creds_config.get_frozen_credentials()
    region = s3_config["profiles"][profile_name]["region_name"]

    s3_config = dict(
        aws_access_key_id=s3_access_key,
        aws_secret_access_key=s3_secret_key,
        region_name=region,
        endpoint_url=endpoint_url,
    )

    return s3_config



from corc.defaults import OCI
from corc.cli.job.job import (
    job_group,
    job_meta_group,
)
from corc.cli.cluster.cluster import (
    cluster_schedule_group,
    select_cluster_group,
    start_cluster_group,
)
from corc.cli.cluster.node import add_node_group
from corc.cli.instance.instance import add_compute_group
from corc.cli.job.job import select_job_group
from corc.cli.job.result import results_group
from corc.cli.network.vcn import add_vcn_group
from corc.cli.network.subnet import add_subnet_group
from corc.cli.platform.oci import add_oci_group
from corc.cli.storage.storage import (
    add_storage_group,
    delete_storage,
    download_storage,
    select_storage,
)
from corc.cli.storage.s3 import add_s3_group, s3_config_group, s3_extra
from corc.cluster import list_clusters, start_cluster, stop_cluster, update_cluster
from corc.job import run, get_results, delete_results, list_results
from corc.instance import launch_instance


def job_cli(parser):
    job_commands = parser.add_subparsers(title="Commands")
    run_parser = job_commands.add_parser("run")
    job_meta_group(run_parser)
    job_group(run_parser)
    select_cluster_group(run_parser)
    cluster_schedule_group(run_parser)
    select_storage(run_parser)
    add_storage_group(run_parser)
    s3_config_group(run_parser)
    add_s3_group(run_parser)
    run_parser.set_defaults(func=run)

    result_parser = job_commands.add_parser("result")
    result_commands = result_parser.add_subparsers(title="Commands")

    get_parser = result_commands.add_parser("get")
    select_job_group(get_parser)
    results_group(get_parser)
    select_storage(get_parser)
    download_storage(get_parser)
    s3_config_group(get_parser)
    get_parser.set_defaults(func=get_results)

    delete_parser = result_commands.add_parser("delete")
    select_job_group(delete_parser)
    results_group(delete_parser)
    select_storage(delete_parser)
    s3_config_group(delete_parser)
    delete_storage(delete_parser)
    delete_parser.set_defaults(func=delete_results)

    list_parser = result_commands.add_parser("list")
    select_job_group(list_parser, required=False)
    select_storage(list_parser)
    s3_config_group(list_parser)
    s3_extra(list_parser)
    list_parser.set_defaults(func=list_results)


def add_cluster_cli(parser):
    cluster_commands = parser.add_subparsers(title="Commands")
    start_parser = cluster_commands.add_parser("start")
    start_cluster_group(start_parser)
    add_node_group(start_parser)
    add_vcn_group(start_parser)
    add_subnet_group(start_parser)
    start_parser.set_defaults(func=start_cluster)

    stop_parser = cluster_commands.add_parser("stop")
    select_cluster_group(stop_parser)
    stop_parser.set_defaults(func=stop_cluster)

    list_parser = cluster_commands.add_parser("list")
    list_parser.set_defaults(func=list_clusters)

    update_parser = cluster_commands.add_parser("update")
    select_cluster_group(update_parser)
    update_parser.set_defaults(func=update_cluster)


def add_instance_cli(parser):
    instance_commands = parser.add_subparsers(title="Commands")
    launch_parser = instance_commands.add_parser("launch")
    add_compute_group(launch_parser)
    add_vcn_group(launch_parser)
    add_subnet_group(launch_parser)
    launch_parser.set_defaults(func=launch_instance)


def add_platform_group(parser):
    platform_group = parser.add_argument_group(title="Available Platforms")
    platform_group.add_argument("platform", choices=[OCI], default=OCI)
    add_oci_group(parser)



from kubernetes import client
from corc.scheduler import Scheduler
from corc.kubernetes.config import load_kube_config
from corc.kubernetes.job import create_job, prepare_job
from corc.kubernetes.storage import (
    prepare_volume,
    prepare_secret,
    create_secret,
    create_volume,
)


class KubenetesScheduler(Scheduler):
    def __init__(self):
        loaded = load_kube_config()
        if not loaded:
            raise RuntimeError("Failed to load the kubernetes config")
        self.batch_client = client.BatchV1Api()
        self.core_client = client.CoreV1Api()

    def prepare(self, **config):
        if not config:
            return False

        if "secret_kwargs" in config:
            secret_prepared = prepare_secret(**config)
            if not secret_prepared:
                return False

            # create secret
            secret_created = create_secret(self.core_client, secret_prepared)
            if not secret_created:
                return False

        if "volume_kwargs" in config:
            volume_prepared = prepare_volume(**config)
            if not volume_prepared:
                return False

            volume_created = create_volume(self.core_client, volume_prepared)
            if not volume_created:
                return False

        return True

    def submit(self, **task):
        if not task:
            return False

        job = prepare_job(**task)
        if not job:
            return False

        # Schedule job on cluster
        job_created = create_job(self.batch_client, job)
        if not job_created:
            return False
        return True

    def retrieve(self, job_id):
        return self.jobs[job_id]



import os
from ansible.parsing.dataloader import DataLoader
from ansible.inventory.manager import InventoryManager
from ansible.vars.manager import VariableManager
from ansible.playbook import Playbook
from ansible.executor.task_queue_manager import TaskQueueManager
from args import get_arguments, OCI, ANSIBLE

here = os.path.dirname(os.path.abspath(__file__))


def load_playbook_source(path):
    loader = DataLoader()
    play_source = loader.load_from_file(path)
    return play_source


def run_playbook(
    play, variable_manager=None, inventory_manager=None, loader=None, passwords=None
):
    tqm = None
    try:
        tqm = TaskQueueManager(
            inventory=inventory_manager,
            variable_manager=variable_manager,
            loader=loader,
            passwords=passwords,
        )
        _ = tqm.run(play)
    finally:
        if tqm is not None:
            tqm.cleanup()
        if loader:
            loader.cleanup_all_tmp_files()


if __name__ == "__main__":
    args = get_arguments([OCI, ANSIBLE], strip_group_prefix=True)
    # Load inventory
    loader = DataLoader()
    loader.set_basedir(args.root_path)
    inventory_manager = InventoryManager(loader, sources=args.inventory_path)
    variable_manager = VariableManager(loader=loader, inventory=inventory_manager)

    # After being started -> run ansible playbook (wagstaff)
    # Load playbook
    pb = Playbook.load(
        args.playbook_path, variable_manager=variable_manager, loader=loader
    )
    plays = pb.get_plays()

    for play in plays:
        run_playbook(
            play,
            variable_manager=variable_manager,
            inventory_manager=inventory_manager,
            loader=loader,
        )



def valid_compute_group(parser):
    add_compute_group(parser)


def add_compute_group(parser):
    compute_group = parser.add_argument_group(title="Compute arguments")
    compute_group.add_argument("--compute-ssh-authorized-keys", nargs="+", default=[])
    compute_group.add_argument("--compute-ad", default="", required=True)
    compute_group.add_argument("--compute-os", default="CentOS")
    compute_group.add_argument("--compute-os-version", default="7")
    compute_group.add_argument("--compute-target-shape", default="VM.Standard2.1")



import argparse
from argparse import Namespace
from corc.defaults import (
    ANSIBLE,
    CLUSTER,
    COMPUTE,
    JOB,
    NODE,
    OCI,
    STORAGE,
    SUBNET,
    S3,
    VCN,
)
from corc.cli.configurer.ansible import valid_ansible_group
from corc.cli.cluster.cluster import valid_cluster_group
from corc.cli.instance.instance import valid_compute_group
from corc.cli.job.job import valid_job_group
from corc.cli.network.subnet import valid_subnet_group
from corc.cli.network.vcn import valid_vcn_group
from corc.cli.cluster.node import valid_node_group
from corc.cli.platform.oci import valid_oci_group
from corc.cli.storage.storage import valid_storage_group
from corc.cli.storage.s3 import valid_s3_group


def strip_argument_prefix(arguments, prefix=""):
    return {k.replace(prefix, ""): v for k, v in arguments.items()}


def _get_arguments(arguments, startswith=""):
    return {k: v for k, v in arguments.items() if k.startswith(startswith)}


argument_groups = {
    ANSIBLE: valid_ansible_group,
    CLUSTER: valid_cluster_group,
    COMPUTE: valid_compute_group,
    JOB: valid_job_group,
    NODE: valid_node_group,
    OCI: valid_oci_group,
    STORAGE: valid_storage_group,
    SUBNET: valid_subnet_group,
    S3: valid_s3_group,
    VCN: valid_vcn_group,
}


def extract_arguments(arguments, argument_types, strip_group_prefix=True):
    if strip_group_prefix:
        stripped_args = {}
        for argument_group in argument_types:
            group_args = _get_arguments(vars(arguments), argument_group.lower())
            group_args = strip_argument_prefix(group_args, argument_group.lower() + "_")
            stripped_args.update(group_args)
        return Namespace(**stripped_args)
    return {}


def get_arguments(argument_types, strip_group_prefix=True, parser=None):
    if not parser:
        parser = argparse.ArgumentParser()

    for argument_group in argument_types:
        if argument_group in argument_groups:
            argument_groups[argument_group](parser)

    args, unknown = parser.parse_known_intermixed_args()
    # args = parser.parse_args()
    if strip_group_prefix:
        stripped_args = {}
        for argument_group in argument_types:
            group_args = _get_arguments(vars(args), argument_group.lower())
            group_args = strip_argument_prefix(group_args, argument_group.lower() + "_")
            stripped_args.update(group_args)
        return Namespace(**stripped_args)
    return args



def valid_cluster_group(parser):
    cluster_schedule_group(parser)
    select_cluster_group(parser)
    start_cluster_group(parser)


def cluster_schedule_group(parser):
    cluster_group = parser.add_argument_group(title="Cluster Runtime arguments")
    cluster_group.add_argument(
        "--cluster-image", default="nielsbohr/mccode-job-runner:latest"
    )


def select_cluster_group(parser):
    cluster_group = parser.add_argument_group(title="Cluster Identity arguments")
    cluster = cluster_group.add_mutually_exclusive_group(required=True)
    cluster.add_argument("--cluster-id", default="")
    cluster.add_argument("--cluster-name", default="")


def start_cluster_group(parser):
    cluster_group = parser.add_argument_group(title="Cluster Start arguments")
    cluster_group.add_argument("--cluster-kubernetes-version", default=None)
    cluster_group.add_argument("--cluster-domain", default="", required=True)



import os
import unittest
from corc.config.config import (
    generate_default_config,
    save_config,
    load_config,
    remove_config
)
from corc.config.defaults import default_config


class ConfigTest(unittest.TestCase):
    def setUp(self):
        self.test_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tmp")
        self.config_path = os.path.join(self.test_dir, "config")

    def tearDown(self):
        self.assertTrue(remove_config(self.config_path))

    def test_generate_config(self):
        self.assertDictEqual(generate_default_config(), default_config)

    def test_save_config(self):
        config = generate_default_config()
        self.assertTrue(save_config(config, path=self.config_path))

    def test_load_config(self):
        config = generate_default_config()
        self.assertTrue(save_config(config, path=self.config_path))
        loaded_config = load_config(self.config_path)
        self.assertIsInstance(loaded_config, dict)
        self.assertDictEqual(config, loaded_config)

    # def test_valid_config(self):
    #     config = generate_default_config()
    #     self.assertTrue(valid_config(config, verbose=True, throw=True))



def valid_storage_group(parser):
    add_storage_group(parser)
    delete_storage(parser)
    download_storage(parser)
    select_storage(parser)


def add_storage_group(parser):
    storage_group = parser.add_argument_group(title="Storage arguments")
    storage_group.add_argument("--storage-credentials-path", default="/mnt/creds")
    storage_group.add_argument("--storage-upload-path", default="")
    storage_group.add_argument("--storage-input-path", default="/tmp/input")
    storage_group.add_argument("--storage-output-path", default="/tmp/output")


def delete_storage(parser):
    delete_storage_group = parser.add_argument_group(title="Storage Delete arguments")
    delete_method = delete_storage_group.add_mutually_exclusive_group()
    delete_method.add_argument(
        "--storage-delete-all", action="store_true", default=False
    )
    delete_method.add_argument("--storage-delete-id", default="")
    delete_method.add_argument("--storage-delete-prefix", default="")


def download_storage(parser):
    get_storage_group = parser.add_argument_group(title="Storage Download arguments")
    get_storage_group.add_argument("--storage-download-path", default="")


def select_storage(parser):
    storage_group = parser.add_argument_group(title="Storage Selection arguments")
    storage_group.add_argument("--storage-endpoint", default="", required=True)
    storage_providers = storage_group.add_mutually_exclusive_group()
    storage_providers.add_argument("--storage-s3", action="store_true", default=True)



import os
import unittest
from corc.oci.cluster import OCIClusterOrchestrator


class TestClusterOrchestrator(unittest.TestCase):
    def setUp(self):
        # Load compartment_id from the env
        if "OCI_COMPARTMENT_ID" not in os.environ:
            raise ValueError("Missing required environment variable OCI_COMPARTMENT_ID")

        if "OCI_PROFILE_NAME" in os.environ:
            profile_name = os.environ["OCI_PROFILE_NAME"]
        else:
            profile_name = "DEFAULT"

        oci_options = dict(
            compartment_id=os.environ["OCI_COMPARTMENT_ID"], profile_name=profile_name,
        )

        test_name = "Test_C_Orch"
        cluster_name = test_name
        node_name = test_name + "_Node"
        vcn_name = test_name + "_Network"
        subnet_name = test_name + "_Subnet"

        # Add unique test postfix
        if "OCI_TEST_ID" in os.environ:
            cluster_name += os.environ["OCI_TEST_ID"]
            node_name += os.environ["OCI_TEST_ID"]
            vcn_name += os.environ["OCI_TEST_ID"]
            subnet_name += os.environ["OCI_TEST_ID"]

        cluster_options = dict(name=cluster_name,)
        image_options = dict(display_name="Oracle-Linux-7.7-2020.03.23-0",)

        node_options = dict(
            availability_domain="lfcb:EU-FRANKFURT-1-AD-1",
            name=node_name,
            size=1,
            node_shape="VM.Standard1.1",
            image=image_options,
        )

        vcn_options = dict(
            cidr_block="10.0.0.0/16", display_name=vcn_name, dns_label="ku",
        )

        subnet_options = dict(display_name=subnet_name, dns_label="workers")

        self.options = dict(
            oci=oci_options,
            cluster=cluster_options,
            node=node_options,
            vcn=vcn_options,
            subnet=subnet_options,
        )

        OCIClusterOrchestrator.validate_options(self.options)
        self.orchestrator = OCIClusterOrchestrator(self.options)
        # Should not be ready at this point
        self.assertFalse(self.orchestrator.is_ready())

    def tearDown(self):
        self.orchestrator.tear_down()
        self.assertFalse(self.orchestrator.is_ready())
        self.orchestrator = None
        self.options = None

    def test_setup_cluster(self):
        self.orchestrator.setup()
        self.assertTrue(self.orchestrator.is_ready())

    def test_teardown_cluster(self):
        self.assertFalse(self.orchestrator.is_ready())
        self.orchestrator.tear_down()
        self.assertFalse(self.orchestrator.is_ready())


if __name__ == "__main__":
    unittest.main()



import os
from corc.util import dump_yaml
from kubernetes import config
from kubernetes.config import ConfigException
from kubernetes.config.kube_config import KUBE_CONFIG_DEFAULT_LOCATION


def load_kube_config():
    try:
        config.load_kube_config()
        return True
    except (ConfigException, TypeError) as err:
        print("Failed to load kube config: {}".format(err))
    return False


def save_kube_config(config_dict, config_file=None, user_exec_args=None):
    if not isinstance(config_dict, dict):
        return False
    if config_file is None:
        # Get the absolute path
        config_file = os.path.expanduser(KUBE_CONFIG_DEFAULT_LOCATION)
    return dump_yaml(config_file, config_dict)



# Package name
PACKAGE_NAME = "corc"

# Argument group defaults
ANSIBLE = "ANSIBLE"

# Cluster
CLUSTER = "CLUSTER"

COMPUTE = "COMPUTE"
JOB = "JOB"
NODE = "NODE"
OCI = "OCI"
RUN = "RUN"
STORAGE = "STORAGE"
SUBNET = "SUBNET"
S3 = "S3"
VCN = "VCN"

# Storage credentials secret name
STORAGE_CREDENTIALS_NAME = "storage-credentials"

# Kubernetes defaults
KUBERNETES_NAMESPACE = "default"

# Job defaults
JOB_DEFAULT_NAME = "job"



def valid_ansible_group(parser):
    add_ansible_group(parser)


def add_ansible_group(parser):
    ansible_group = parser.add_argument_group(title="Ansible arguments")
    ansible_group.add_argument("--ansible-root-path", default=False)
    ansible_group.add_argument("--ansible-playbook-path", default=False)
    ansible_group.add_argument("--ansible-inventory-path", default=False)



from corc.defaults import CLUSTER, OCI, JOB, S3, STORAGE
from corc.cli.args import extract_arguments
from corc.oci.job import (
    run as oci_run,
    get_results as oci_get_results,
    delete_results as oci_delete_results,
    list_results as oci_list_results,
)


def run(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    cluster_args = vars(extract_arguments(args, [CLUSTER]))
    job_args = vars(extract_arguments(args, [JOB]))
    staging_args = vars(extract_arguments(args, [STORAGE]))
    s3_args = vars(extract_arguments(args, [S3]))

    if not s3_args["bucket_name"]:
        s3_args["bucket_name"] = job_args["name"]

    if oci_args:
        oci_run(
            cluster_kwargs=cluster_args,
            job_kwargs=job_args,
            oci_kwargs=oci_args,
            staging_kwargs=staging_args,
            storage_kwargs=s3_args,
        )


def get_results(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    job_args = vars(extract_arguments(args, [JOB]))
    staging_args = vars(extract_arguments(args, [STORAGE]))
    s3_args = vars(extract_arguments(args, [S3]))

    if oci_args:
        return oci_get_results(
            job_kwargs=job_args, staging_kwargs=staging_args, storage_kwargs=s3_args,
        )


def delete_results(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    job_args = vars(extract_arguments(args, [JOB]))
    staging_args = vars(extract_arguments(args, [STORAGE]))
    s3_args = vars(extract_arguments(args, [S3]))

    if oci_args:
        return oci_delete_results(
            job_kwargs=job_args, staging_kwargs=staging_args, storage_kwargs=s3_args
        )


def list_results(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    job_args = vars(extract_arguments(args, [JOB]))
    staging_args = vars(extract_arguments(args, [STORAGE]))
    s3_args = vars(extract_arguments(args, [S3]))

    if oci_args:
        if "extra_kwargs" in s3_args:
            storage_extra_kwargs = s3_args.pop("extra_kwargs")
        return oci_list_results(
            job_kwargs=job_args,
            staging_kwargs=staging_args,
            storage_kwargs=s3_args,
            storage_extra_kwargs=storage_extra_kwargs,
        )



from kubernetes import client


def prepare_secret(secret_kwargs=None):
    if not secret_kwargs:
        secret_kwargs = {}

    secret = client.V1Secret(**secret_kwargs)
    return secret


def create_secret(api_instance, secret, namespace="default"):
    api_response = api_instance.create_namespaced_secret(
        body=secret, namespace=namespace
    )
    return api_response


def prepare_volume(volume_kwargs=None):
    if not volume_kwargs:
        volume_kwargs = {}

    volume = client.V1Volume(**volume_kwargs)
    return volume


def create_volume(api_instance, volume, namespace="default"):
    api_response = api_instance.create_namespaced_volume(
        body=volume, namespace=namespace
    )
    return api_response


class StorageProvider:
    def __init__(self, config):
        self.config = config
        self.core_api = client.CoreV1Api()

    def provision(self):
        pass
        # Provision secrets
        # for

        # Provision volumes



import oci
from oci.core import (
    ComputeClient,
    ComputeClientCompositeOperations,
    VirtualNetworkClient,
    VirtualNetworkClientCompositeOperations,
)
from oci.identity import IdentityClient, IdentityClientCompositeOperations
from oci.core.models import (
    InstanceSourceViaImageDetails,
    LaunchInstanceDetails,
    CreateVnicDetails,
    Instance,
)
from corc.orchestrator import Orchestrator
from corc.util import open_port
from corc.oci.helpers import (
    create,
    delete,
    get,
    list_entities,
    new_client,
    stack_was_deleted,
)
from corc.oci.network import (
    get_vcn_by_name,
    new_vcn_stack,
    valid_vcn_stack,
    get_vcn_stack,
    get_subnet_by_name,
    delete_vcn_stack,
    refresh_vcn_stack,
)


def valid_instance(instance):
    if not isinstance(instance, Instance):
        return False
    return True


def create_instance(
    compute_client, launch_instance_details, wait_for_states=None, **kwargs
):
    if not wait_for_states:
        wait_for_states = [oci.core.models.Instance.LIFECYCLE_STATE_RUNNING]

    instance = create(
        compute_client,
        "launch_instance",
        launch_instance_details,
        wait_for_states=wait_for_states,
        **kwargs,
    )

    if not instance:
        return None
    return instance


def terminate_instance(compute_client, instance_id, **kwargs):
    return delete(compute_client, "terminate_instance", instance_id, **kwargs)


def get_instance_by_name(compute_client, compartment_id, display_name, kwargs=None):
    if not kwargs:
        kwargs = {}

    if "lifecycle_state" not in kwargs:
        kwargs.update(dict(lifecycle_state=Instance.LIFECYCLE_STATE_RUNNING))

    instances = list_entities(
        compute_client,
        "list_instances",
        compartment_id,
        display_name=display_name,
        **kwargs,
    )
    if instances:
        return instances[0]
    return None


def launch_instance(compute_kwargs={}, oci_kwargs={}, subnet_kwargs={}, vcn_kwargs={}):
    raise NotImplementedError
    # compute_client = new_client(
    #     ComputeClient,
    #     composite_class=ComputeClientCompositeOperations,
    #     profile_name=oci_kwargs["profile_name"],
    # )

    # identity_client = new_client(
    #     IdentityClient,
    #     composite_class=IdentityClientCompositeOperations,
    #     profile_name=oci_kwargs["profile_name"],
    # )

    # network_client = new_client(
    #     VirtualNetworkClient,
    #     composite_class=VirtualNetworkClientCompositeOperations,
    #     profile_name=oci_kwargs["profile_name"],
    # )

    # # AD
    # selected_availability_domain = None
    # available_domains = list_entities(
    #     identity_client,
    #     "list_availability_domains",
    #     compartment_id=oci_kwargs["compartment_id"],
    # )

    # for domain in available_domains:
    #     if domain.name == compute_kwargs["ad"]:
    #         selected_availability_domain = domain

    # if not selected_availability_domain:
    #     print("Failed to find the selected AD: {}".format(compute_kwargs["ad"]))
    #     print("Available are: {}".format(available_domains))
    #     return False

    # # Available images
    # available_images = list_entities(
    #     compute_client, "list_images", oci_kwargs["compartment_id"]
    # )

    # available_shapes = list_entities(
    #     compute_client, "list_shapes", oci_kwargs["compartment_id"]
    # )

    # instance_details = _gen_instance_stack_details(
    #     self.vcn_stack["vcn"].id,
    #     subnet.id,
    #     available_images,
    #     available_shapes,
    #     self.options,
    # )

    # # Network (VCN)
    # selected_vcn = None
    # get_vcn_by_name(
    #     network_client, oci_kwargs["compartment_id"],
    # )

    # # VCN Subnet
    # selected_subnet = None
    # if not selected_subnet:
    #     print("Failed to find the selected subnet: {}".format(subnet_kwargs["id"]))
    #     return False

    # # Existing or create new vnic
    # create_vnic_details = CreateVnicDetails(subnet_id=selected_subnet.id)

    # launch_instance_details = LaunchInstanceDetails(
    #     compartment_id=oci_kwargs["compartment_id"],
    #     availability_domain=selected_availability_domain.name,
    #     shape=selected_shape.shape,
    # )

    # instance = create_instance(
    #     compute_client,
    #     launch_instance_details,
    #     wait_for_states=[oci.core.models.Instance.LIFECYCLE_STATE_PROVISIONING],
    # )

    # print("Instance: {}".format(instance))


def _prepare_source_details(available_images, **kwargs):
    selected_image = None
    for image in available_images:
        if (
            image.operating_system == kwargs["operating_system"]
            and image.operating_system_version == kwargs["operating_system_version"]
        ):
            selected_image = image
            break

    if not selected_image:
        raise RuntimeError("Failed to find the specified image")

    source_image_kwargs = {
        k: v for k, v in kwargs.items() if hasattr(InstanceSourceViaImageDetails, k)
    }
    source_image_details = InstanceSourceViaImageDetails(
        image_id=selected_image.id, **source_image_kwargs
    )
    return source_image_details


def _prepare_shape(available_shapes, **kwargs):
    selected_shape = None
    for shape in available_shapes:
        if shape.shape == kwargs["shape"]:
            selected_shape = shape.shape
            break

    if not selected_shape:
        raise RuntimeError("Failed to find the specified shape")

    return selected_shape


def _prepare_vnic_details(**kwargs):
    create_nvic_kwargs = {
        k: v for k, v in kwargs.items() if hasattr(CreateVnicDetails, k)
    }
    create_vnic_details = CreateVnicDetails(**create_nvic_kwargs)
    return create_vnic_details


def _prepare_metadata(**kwargs):
    prepared_metadata = {}
    if "ssh_authorized_keys" in kwargs:
        prepared_metadata.update(
            {"ssh_authorized_keys": "\n".join(kwargs["ssh_authorized_keys"])}
        )
        kwargs.pop("ssh_authorized_keys")

    prepared_metadata.update(**kwargs)
    return prepared_metadata


def _prepare_launch_instance_details(**kwargs):
    instance_kwargs = {
        k: v for k, v in kwargs.items() if hasattr(LaunchInstanceDetails, k)
    }
    launch_instance_details = LaunchInstanceDetails(**instance_kwargs)
    return launch_instance_details


def _gen_instance_stack_details(vcn_id, subnet_id, images, shapes, **options):
    instance_stack_details = {}

    source_details = _prepare_source_details(images, **options["compute"])

    shape = _prepare_shape(shapes, **options["compute"])

    create_vnic_details = _prepare_vnic_details(subnet_id=subnet_id)

    # Extract metadata
    metadata = {}
    if "compute_metadata" in options:
        metadata = _prepare_metadata(**options["compute_metadata"])

    if "shape" in options["compute"]:
        options["compute"].pop("shape")

    launch_instance_details = _prepare_launch_instance_details(
        compartment_id=options["oci"]["compartment_id"],
        shape=shape,
        create_vnic_details=create_vnic_details,
        source_details=source_details,
        metadata=metadata,
        **options["compute"],
    )

    instance_stack_details["launch_instance"] = launch_instance_details
    return instance_stack_details


class OCIInstanceOrchestrator(Orchestrator):
    def __init__(self, options):
        super().__init__(options)

        self.compute_client = new_client(
            ComputeClient,
            composite_class=ComputeClientCompositeOperations,
            profile_name=options["oci"]["profile_name"],
        )

        self.identity_client = new_client(
            IdentityClient,
            composite_class=IdentityClientCompositeOperations,
            profile_name=options["oci"]["profile_name"],
        )

        self.network_client = new_client(
            VirtualNetworkClient,
            composite_class=VirtualNetworkClientCompositeOperations,
            profile_name=options["oci"]["profile_name"],
        )

        self.port = 22
        self.instance = None
        self.vcn_stack = None
        self._is_ready = False

    def _get_vcn_stack(self):
        stack = {}
        vcn = get_vcn_by_name(
            self.network_client,
            self.options["oci"]["compartment_id"],
            self.options["vcn"]["display_name"],
        )
        if vcn:
            stack = get_vcn_stack(
                self.network_client, self.options["oci"]["compartment_id"], vcn.id
            )
        return stack

    def _new_vcn_stack(self):
        stack = new_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )
        return stack

    def _refresh_vcn_stack(self, vcn_stack):
        stack = refresh_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )
        return stack

    def endpoint(self, select=None):
        # Return the endpoint that is being orchestrated
        public_endpoint = None
        if self.instance:
            vnic_attachments = list_entities(
                self.compute_client,
                "list_vnic_attachments",
                self.options["oci"]["compartment_id"],
                instance_id=self.instance.id,
            )
            # For now just pick the first attachment
            for vnic_attach in vnic_attachments:
                vnic = get(self.network_client, "get_vnic", vnic_attach.vnic_id)
                if hasattr(vnic, "public_ip") and vnic.public_ip:
                    public_endpoint = vnic.public_ip
                    break

        return public_endpoint

    def setup(self):
        # Ensure we have a VCN stack ready
        vcn_stack = self._get_vcn_stack()
        if not vcn_stack:
            vcn_stack = self._new_vcn_stack()

        if valid_vcn_stack(vcn_stack):
            self.vcn_stack = vcn_stack
        else:
            # FIXME, creates a duplicate VCN
            self.vcn_stack = self._refresh_vcn_stack(vcn_stack)

        if not valid_vcn_stack(self.vcn_stack):
            raise RuntimeError(
                "Failed to fix the broken VCN stack: {}".format(vcn_stack)
            )

        # Find the selected subnet in the VCN
        subnet = get_subnet_by_name(
            self.network_client,
            self.options["oci"]["compartment_id"],
            self.vcn_stack["id"],
            self.options["subnet"]["display_name"],
        )

        if not subnet:
            raise RuntimeError(
                "Failed to find a subnet with the name: {} in vcn: {}".format(
                    self.options["subnet"]["display_name"], self.vcn_stack["vcn"].id
                )
            )

        # Available images
        available_images = list_entities(
            self.compute_client, "list_images", self.options["oci"]["compartment_id"]
        )

        available_shapes = list_entities(
            self.compute_client, "list_shapes", self.options["oci"]["compartment_id"]
        )

        instance_details = _gen_instance_stack_details(
            self.vcn_stack["vcn"].id,
            subnet.id,
            available_images,
            available_shapes,
            **self.options,
        )

        instance = None
        if "display_name" in self.options["compute"]:
            instance = get_instance_by_name(
                self.compute_client,
                self.options["oci"]["compartment_id"],
                self.options["compute"]["display_name"],
            )

        if not instance:
            self._is_ready = False
            instance = create_instance(
                self.compute_client, instance_details["launch_instance"]
            )
            if valid_instance(instance):
                self.instance = instance
            else:
                raise ValueError("The new instance: {} is not valid".format(instance))
        else:
            if valid_instance(instance):
                self.instance = instance

        if self.instance:
            self._is_ready = True

    def tear_down(self):
        if not self.instance:
            self.instance = get_instance_by_name(
                self.compute_client,
                self.options["oci"]["compartment_id"],
                self.options["compute"]["display_name"],
            )

        if self.instance:
            # Await that it is terminated
            # Must be done before we can remove the vcn
            deleted = terminate_instance(
                self.compute_client,
                self.instance.id,
                wait_for_states=[Instance.LIFECYCLE_STATE_TERMINATED],
            )
            if deleted:
                self.instance = None
        else:
            self.instance = None

        if not self.vcn_stack:
            # refresh
            self.vcn_stack = self._get_vcn_stack()

        # TODO, optional delete
        if self.vcn_stack:
            vcn_deleted = delete_vcn_stack(
                self.network_client,
                self.options["oci"]["compartment_id"],
                vcn_id=self.vcn_stack["id"],
            )
            if stack_was_deleted(vcn_deleted):
                self.vcn_stack = None

        if self.instance and self.vcn_stack:
            self._is_ready = True
        else:
            self._is_ready = False

    def poll(self):
        target_endpoint = self.endpoint()
        if target_endpoint:
            if open_port(target_endpoint, self.port):
                self._is_reachable = True

    @classmethod
    def validate_options(cls, options):
        if not isinstance(options, dict):
            raise ValueError("options is not a dictionary")

        expected_oci_keys = [
            "compartment_id",
            "profile_name",
        ]

        expected_compute_keys = [
            "availability_domain",
            "shape",
            "operating_system",
            "operating_system_version",
        ]

        optional_compute_metadata_keys = ["ssh_authorized_keys"]
        optional_compute_keys = ["display_name"]

        expected_vcn_keys = ["cidr_block", "dns_label", "display_name"]

        optional_vcn_keys = ["id"]

        expected_subnet_keys = ["dns_label"]

        optional_subnet_keys = ["id", "cidr_block", "display_name"]

        expected_groups = {
            "oci": expected_oci_keys,
            "compute": expected_compute_keys
            + optional_compute_metadata_keys
            + optional_compute_keys,
            "vcn": expected_vcn_keys + optional_vcn_keys,
            "subnet": expected_subnet_keys + optional_subnet_keys,
        }

        for group, keys in expected_groups.items():
            if group not in options:
                raise KeyError("Missing group: {}".format(group))

            if not isinstance(options[group], dict):
                raise TypeError("Group: {} must be a dictionary".format(group))

            for key, _ in options[group].items():
                if key not in keys:
                    raise KeyError("Incorrect key: {} is not in: {}".format(key, keys))



from ansible.parsing.dataloader import DataLoader
from ansible.inventory.manager import InventoryManager
from ansible.vars.manager import VariableManager
from ansible.playbook import Playbook
from ansible.executor.task_queue_manager import TaskQueueManager


def load_playbook_source(path):
    loader = DataLoader()
    play_source = loader.load_from_file(path)
    return play_source


def run_playbook(
    play, variable_manager=None, inventory_manager=None, loader=None, passwords=None
):
    tqm = None
    try:
        tqm = TaskQueueManager(
            inventory=inventory_manager,
            variable_manager=variable_manager,
            loader=loader,
            passwords=passwords,
        )
        _ = tqm.run(play)
    finally:
        if tqm is not None:
            tqm.cleanup()
        if loader:
            loader.cleanup_all_tmp_files()


class AnsibleConfigurer:
    def __init__(self, options):
        self.options = options
        self.loader = DataLoader()
        if "inventory_path" in self.options:
            sources = self.options["inventory_path"]
        else:
            sources = None

        self.inventory_manager = InventoryManager(self.loader, sources=sources)
        self.inventory_manager.add_group("compute")

        self.variable_manager = VariableManager(
            loader=self.loader, inventory=self.inventory_manager
        )
        for host in self.options["hosts"]:
            self.inventory_manager.add_host(host, group="compute", port="22")

        self.variable_manager.set_host_variable(host, "ansible_connection", "ssh")
        self.variable_manager.set_host_variable(host, "ansible_user", "opc")
        self.variable_manager.set_host_variable(host, "ansible_become", "yes")
        self.variable_manager.set_host_variable(host, "ansible_become_method", "sudo")
        self.variable_manager.set_host_variable(
            host, "ansible_host_key_checking", "False"
        )

        # Use the provided ssh key
        self.variable_manager.set_host_variable(
            host, "ansible_ssh_private_key_file", self.options["ssh_private_key_file"]
        )

        self.variable_manager.set_host_variable(host, "verbosity", 4)

    def apply(self):
        playbook_path = self.options["playbook_path"]
        playbook = Playbook.load(
            playbook_path, variable_manager=self.variable_manager, loader=self.loader
        )

        plays = playbook.get_plays()
        for play in plays:
            run_playbook(
                play,
                variable_manager=self.variable_manager,
                inventory_manager=self.inventory_manager,
                loader=self.loader,
            )

    @classmethod
    def validate_options(cls, options):
        if not isinstance(options, dict):
            raise TypeError("options is not a dictionary")

        # expected_ansible_keys = ["inventory_path", "playbook_path"]



def results_group(parser):
    results_group = parser.add_argument_group(title="Job Result arguments")
    results_group.add_argument("--job-result-all", action="store_true", default=True)
    results_group.add_argument("--job-result-prefix", default="")



import os
import unittest
from oci.core import VirtualNetworkClient, VirtualNetworkClientCompositeOperations
from corc.oci.helpers import new_client, get, list_entities, stack_was_deleted
from corc.oci.network import (
    new_vcn_stack,
    delete_vcn_stack,
    valid_vcn_stack,
    get_vcn_stack,
    get_vcn_by_name,
    equal_vcn_stack,
    refresh_vcn_stack,
)


class TestVCNStack(unittest.TestCase):
    def setUp(self):
        # Load compartment_id from the env
        if "OCI_COMPARTMENT_ID" not in os.environ:
            raise ValueError("Missing required environment variable OCI_COMPARTMENT_ID")

        if "OCI_PROFILE_NAME" in os.environ:
            profile_name = os.environ["OCI_PROFILE_NAME"]
        else:
            profile_name = "DEFAULT"

        self.oci_options = dict(
            compartment_id=os.environ["OCI_COMPARTMENT_ID"], profile_name=profile_name,
        )

        test_name = "Test_VCN"
        vcn_name = test_name + "_Network"
        subnet_name = test_name + "_Subnet"

        # Add unique test postfix
        if "OCI_TEST_ID" in os.environ:
            vcn_name += os.environ["OCI_TEST_ID"]
            subnet_name += os.environ["OCI_TEST_ID"]

        self.vcn_options = dict(
            cidr_block="10.0.0.0/16", display_name=vcn_name, dns_label="ku",
        )

        self.subnet_options = dict(display_name=subnet_name, dns_label="workers")

        self.options = dict(
            oci=self.oci_options, vcn=self.vcn_options, subnet=self.subnet_options,
        )

        self.network_client = new_client(
            VirtualNetworkClient,
            composite_class=VirtualNetworkClientCompositeOperations,
            profile_name=self.options["oci"]["profile_name"],
        )

    def tearDown(self):
        # Delete all vcn with display_name
        vcns = list_entities(
            self.network_client,
            "list_vcns",
            self.options["oci"]["compartment_id"],
            display_name=self.vcn_options["display_name"],
        )

        for vcn in vcns:
            deleted_stack = delete_vcn_stack(
                self.network_client, self.options["oci"]["compartment_id"], vcn=vcn
            )
            self.assertTrue(stack_was_deleted(deleted_stack))

    def test_vcn_stack(self):
        # Extract the ip of the instance
        self.vcn_stack = new_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )

        self.assertTrue(valid_vcn_stack(self.vcn_stack))
        _vcn_stack = get_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            self.vcn_stack["id"],
        )
        self.assertTrue(valid_vcn_stack(_vcn_stack))
        self.assertEqual(self.vcn_stack["id"], _vcn_stack["id"])
        self.assertEqual(self.vcn_stack["vcn"], _vcn_stack["vcn"])
        self.assertListEqual(
            self.vcn_stack["internet_gateways"], _vcn_stack["internet_gateways"]
        )
        self.assertListEqual(self.vcn_stack["subnets"], _vcn_stack["subnets"])

        new_vcn_id = get(self.network_client, "get_vcn", self.vcn_stack["id"])

        self.assertEqual(self.vcn_stack["id"], new_vcn_id.id)
        self.assertEqual(self.vcn_stack["vcn"], new_vcn_id)

        # Custom equal check
        self.assertTrue(equal_vcn_stack(self.vcn_stack, _vcn_stack))

        new_vcn_name = get_vcn_by_name(
            self.network_client,
            self.options["oci"]["compartment_id"],
            self.options["vcn"]["display_name"],
        )

        self.assertEqual(self.vcn_stack["id"], new_vcn_name.id)
        self.assertEqual(self.vcn_stack["vcn"], new_vcn_name)

    def test_vcn_delete_stack(self):
        # Extract the ip of the instance
        self.vcn_stack = new_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )

        self.assertTrue(valid_vcn_stack(self.vcn_stack))

        created_vcn_stack = get_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            self.vcn_stack["id"],
        )

        self.assertTrue(valid_vcn_stack(created_vcn_stack))
        self.assertTrue(equal_vcn_stack(self.vcn_stack, created_vcn_stack))

        deleted_stack = delete_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            self.vcn_stack["id"],
        )
        self.assertTrue(stack_was_deleted(deleted_stack))

    def test_vcn_refresh_stack(self):
        # Extract the ip of the instance
        self.vcn_stack = new_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )

        self.assertTrue(valid_vcn_stack(self.vcn_stack))
        refresh_stack = {"id": self.vcn_stack["id"]}

        refreshed_vcn = refresh_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=refresh_stack,
        )
        self.assertTrue(valid_vcn_stack(refreshed_vcn))
        self.assertTrue(equal_vcn_stack(self.vcn_stack, refreshed_vcn))

        refresh_stack = {"display_name": self.vcn_stack["vcn"].display_name}

        refreshed_vcn = refresh_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=refresh_stack,
        )
        self.assertTrue(valid_vcn_stack(refreshed_vcn))
        self.assertTrue(equal_vcn_stack(self.vcn_stack, refreshed_vcn))


if __name__ == "__main__":
    unittest.main()



def valid_node_group(parser):
    add_node_group(parser)
    select_node_group(parser)


def select_node_group(parser):
    node_group = parser.add_argument_group(title="Node Selection arguments")
    node = node_group.add_mutually_exclusive_group(required=True)
    node.add_argument("--node-id", default="")
    node.add_argument("--node-name", default="NodePool")


def add_node_group(parser):
    node_group = parser.add_argument_group(title="Node arguments")
    node_group.add_argument("--node-availability-domain", default="", required=True)
    node_group.add_argument("--node-size", default=1, type=int)
    node_group.add_argument("--node-shape", default="VM.Standard2.1")
    # https://oracle-cloud-infrastructure-python-sdk.readthedocs.io/en/latest/api/cims/models/oci.cims.models.CreateResourceDetails.html?highlight=availability_domain#oci.cims.models.CreateResourceDetails.availability_domain
    node_group.add_argument("--node-image", default="Oracle-Linux-7.7-2020.03.23-0")



import oci
from oci.core import VirtualNetworkClient, VirtualNetworkClientCompositeOperations
from oci.core.models import CreateVcnDetails, Vcn
from oci.core.models import RouteTable, CreateRouteTableDetails, UpdateRouteTableDetails
from oci.core.models import InternetGateway
from oci.core.models import SecurityList
from oci.core.models import DhcpOptions
from oci.core.models import LocalPeeringGateway
from oci.core.models import NatGateway
from oci.core.models import ServiceGateway
from oci.core.models import Subnet, CreateSubnetDetails
from corc.cli.args import get_arguments, OCI, SUBNET, VCN
from corc.oci.helpers import (
    new_client,
    prepare_route_rule,
    list_entities,
    get,
    create,
    delete,
    update,
    stack_was_deleted,
)


def new_vcn_stack(
    network_client, compartment_id, vcn_kwargs=None, subnet_kwargs=None,
):
    if not vcn_kwargs:
        vcn_kwargs = {}

    if "cidr_block" not in vcn_kwargs:
        vcn_kwargs.update(dict(cidr_block="10.0.0.0/16"))

    if "id" in vcn_kwargs:
        vcn_kwargs.pop("id")

    if not subnet_kwargs:
        subnet_kwargs = {}

    if "id" in subnet_kwargs:
        subnet_kwargs.pop("id")

    if "cidr_block" not in subnet_kwargs:
        subnet_kwargs.update(dict(cidr_block="10.0.1.0/24"))

    stack = dict(id=None, vcn=None, internet_gateways=[], subnets=[])
    create_vcn_details = CreateVcnDetails(compartment_id=compartment_id, **vcn_kwargs)
    vcn = create(
        network_client,
        "create_vcn",
        wait_for_states=[Vcn.LIFECYCLE_STATE_AVAILABLE],
        create_vcn_details=create_vcn_details,
    )
    if not vcn:
        raise RuntimeError("Failed to create a vcn with details: {}".format(vcn_kwargs))

    stack["id"] = vcn.id
    stack["vcn"] = vcn

    create_ig_details = oci.core.models.CreateInternetGatewayDetails(
        compartment_id=compartment_id, is_enabled=True, vcn_id=vcn.id
    )
    gateway = create(
        network_client,
        "create_internet_gateway",
        wait_for_states=[InternetGateway.LIFECYCLE_STATE_AVAILABLE],
        create_internet_gateway_details=create_ig_details,
    )
    if not gateway:
        raise RuntimeError(
            "Failed to create internet gateway with details: {}".format(
                create_ig_details
            )
        )

    stack["internet_gateways"].append(gateway)

    # Setup the route table
    route_rule = prepare_route_rule(
        gateway.id,
        cidr_block=None,
        destination="0.0.0.0/0",
        destination_type="CIDR_BLOCK",
    )

    route_rules = []
    if route_rule:
        route_rules.append(route_rule)

    create_rt_details = CreateRouteTableDetails(
        compartment_id=compartment_id, route_rules=route_rules, vcn_id=vcn.id
    )
    route_table = create(
        network_client,
        "create_route_table",
        wait_for_states=[RouteTable.LIFECYCLE_STATE_AVAILABLE],
        create_route_table_details=create_rt_details,
    )

    if not route_table:
        raise RuntimeError(
            "Failed to create route table with details: {}".format(create_rt_details)
        )

    # Create subnet
    create_subnet_details = CreateSubnetDetails(
        compartment_id=compartment_id,
        vcn_id=vcn.id,
        route_table_id=route_table.id,
        **subnet_kwargs
    )
    subnet = create(
        network_client,
        "create_subnet",
        wait_for_states=[Subnet.LIFECYCLE_STATE_AVAILABLE],
        create_subnet_details=create_subnet_details,
    )

    if subnet:
        stack["subnets"].append(subnet)

    return stack


def refresh_vcn_stack(
    network_client,
    compartment_id,
    vcn_kwargs=None,
    gateway_kwargs=None,
    subnet_kwargs=None,
):
    if not vcn_kwargs:
        vcn_kwargs = {}

    if "cidr_block" not in vcn_kwargs:
        vcn_kwargs.update(dict(cidr_block="10.0.0.0/16"))

    if not gateway_kwargs:
        gateway_kwargs = {}

    if not subnet_kwargs:
        subnet_kwargs = {}

    if "cidr_block" not in subnet_kwargs:
        subnet_kwargs.update(dict(cidr_block="10.0.1.0/24"))

    stack = dict(id=None, vcn=None, internet_gateways=[], subnets=[])

    vcn = None
    if "id" in vcn_kwargs:
        vcn = get(network_client, "get_vcn", vcn_kwargs["id"])
    elif "display_name" in vcn_kwargs:
        vcn = get_vcn_by_name(
            network_client, compartment_id, vcn_kwargs["display_name"]
        )

    if not vcn:
        create_vcn_details = CreateVcnDetails(
            compartment_id=compartment_id, **vcn_kwargs
        )
        vcn = create(
            network_client,
            "create_vcn",
            wait_for_states=[Vcn.LIFECYCLE_STATE_AVAILABLE],
            create_vcn_details=create_vcn_details,
        )
        if not vcn:
            raise RuntimeError(
                "Failed to create a vcn with details: {}".format(vcn_kwargs)
            )

    if not vcn:
        raise RuntimeError("Failed to either retrieve or create a vcn")

    stack["id"] = vcn.id
    stack["vcn"] = vcn

    gateways = list_entities(
        network_client,
        "list_internet_gateways",
        compartment_id,
        vcn.id,
        **gateway_kwargs
    )

    if gateways:
        stack["internet_gateways"] = gateways
    else:
        # Create one
        create_ig_details = oci.core.models.CreateInternetGatewayDetails(
            compartment_id=compartment_id, is_enabled=True, vcn_id=vcn.id
        )
        gateway = create(
            network_client,
            "create_internet_gateway",
            wait_for_states=[InternetGateway.LIFECYCLE_STATE_AVAILABLE],
            create_internet_gateway_details=create_ig_details,
        )
        if not gateway:
            raise RuntimeError(
                "Failed to create internet gateway with details: {}".format(
                    create_ig_details
                )
            )
        stack["internet_gateways"].append(gateway)

    subnets = list_entities(network_client, "list_subnets", compartment_id, vcn.id)
    if subnets:
        stack["subnets"] = subnets
    else:
        # Setup the default route table
        route_rules = []
        for gateway in stack["internet_gateways"]:
            route_rule = prepare_route_rule(
                gateway.id,
                cidr_block=None,
                destination="0.0.0.0/0",
                destination_type="CIDR_BLOCK",
            )

            if route_rule:
                route_rules.append(route_rule)

        create_rt_details = CreateRouteTableDetails(
            compartment_id=compartment_id, route_rules=route_rules, vcn_id=vcn.id
        )
        route_table = create(
            network_client,
            "create_route_table",
            wait_for_states=[RouteTable.LIFECYCLE_STATE_AVAILABLE],
            create_route_table_details=create_rt_details,
        )

        # Create new subnet
        create_subnet_details = CreateSubnetDetails(
            compartment_id=compartment_id,
            vcn_id=vcn.id,
            route_table_id=route_table.id,
            **subnet_kwargs
        )
        subnet = create(
            network_client,
            "create_subnet",
            wait_for_states=[Subnet.LIFECYCLE_STATE_AVAILABLE],
            create_subnet_details=create_subnet_details,
        )
        if subnet:
            stack["subnets"].append(subnet)

    return stack


def valid_vcn_stack(stack):

    if not isinstance(stack, dict):
        raise TypeError("The VCN stack must be a dictionary to be valid")

    expected_fields = ["id", "vcn", "internet_gateways", "subnets"]
    for field in expected_fields:
        if field not in stack:
            return False

        if not stack[field]:
            return False
    return True


def get_vcn_stack(
    network_client, compartment_id, vcn_id, subnet_kwargs=None, gateway_kwargs=None
):

    if not subnet_kwargs:
        subnet_kwargs = dict()

    if not gateway_kwargs:
        gateway_kwargs = dict()

    stack = {}
    vcn = get(network_client, "get_vcn", vcn_id)
    if not vcn:
        return stack
    gateways = list_entities(
        network_client,
        "list_internet_gateways",
        compartment_id,
        vcn.id,
        **gateway_kwargs
    )
    subnets = list_entities(
        network_client, "list_subnets", compartment_id, vcn.id, **subnet_kwargs
    )
    stack = {
        "id": vcn.id,
        "vcn": vcn,
        "internet_gateways": gateways,
        "subnets": subnets,
    }
    return stack


def equal_vcn_stack(vcn_stack, other_vcn_stack):
    if vcn_stack["id"] != other_vcn_stack["id"]:
        return False

    if vcn_stack["vcn"] != other_vcn_stack["vcn"]:
        return False

    if len(vcn_stack["internet_gateways"]) != len(other_vcn_stack["internet_gateways"]):
        return False

    for i, gateway in enumerate(vcn_stack["internet_gateways"]):
        if gateway != other_vcn_stack["internet_gateways"][i]:
            return False

    if len(vcn_stack["subnets"]) != len(vcn_stack["subnets"]):
        return False

    for i, subnet in enumerate(vcn_stack["subnets"]):
        if subnet != other_vcn_stack["subnets"][i]:
            return False

    return True


def get_vcn_by_name(network_client, compartment_id, display_name, **kwargs):
    vcns = list_entities(network_client, "list_vcns", compartment_id, **kwargs)
    for vcn in vcns:
        if vcn.display_name == display_name:
            return vcn


def delete_vcn_stack(network_client, compartment_id, vcn_id=None, vcn=None):
    if not vcn_id and not vcn:
        raise ValueError("Either vcn_id or vcn must be provided")

    if vcn_id:
        vcn = get(network_client, "get_vcn", vcn_id)

    remove_stack = {
        "id": False,
        "vcn": False,
        "subnets": [],
        "route_tables": [],
        "internet_gateways": [],
        "security_lists": [],
        "dhcp_options": [],
        "local_peering_gateways": [],
        "nat_gateways": [],
        "service_gateways": [],
    }
    if vcn:
        vcn_subnets = list_entities(
            network_client, "list_subnets", compartment_id, vcn.id
        )
        for subnet in vcn_subnets:
            deleted = delete(
                network_client,
                "delete_subnet",
                subnet.id,
                wait_for_states=[Subnet.LIFECYCLE_STATE_TERMINATED],
            )
            remove_stack["subnets"].append(deleted)

        # Delete all the routes (and disable the gateway target
        # if they are the default which means that they can't be deleted)
        routes = list_entities(
            network_client,
            "list_route_tables",
            compartment_id,
            vcn.id,
            sort_by="TIMECREATED",
            sort_order="ASC",
        )
        # Disable routes on the default route table
        if routes:
            # Disable all routes
            for route in routes:
                update_details = UpdateRouteTableDetails(route_rules=[])
                update(
                    network_client,
                    "update_route_table",
                    route.id,
                    wait_for_states=[RouteTable.LIFECYCLE_STATE_AVAILABLE],
                    update_route_table_details=update_details,
                )

            # Delete all non default routes
            if len(routes) > 1:
                for route in routes[1:]:
                    deleted = delete(
                        network_client,
                        "delete_route_table",
                        route.id,
                        wait_for_states=[RouteTable.LIFECYCLE_STATE_TERMINATED],
                    )
                    remove_stack["route_tables"].append(deleted)

        # Delete all gateways (can't delete the default)
        gateways = list_entities(
            network_client,
            "list_internet_gateways",
            compartment_id,
            vcn.id,
            sort_by="TIMECREATED",
            sort_order="ASC",
        )
        for gateway in gateways:
            deleted = delete(
                network_client,
                "delete_internet_gateway",
                gateway.id,
                wait_for_states=[InternetGateway.LIFECYCLE_STATE_TERMINATED],
            )
            remove_stack["internet_gateways"].append(deleted)

        # Delete all security lists
        securities = list_entities(
            network_client,
            "list_security_lists",
            compartment_id,
            vcn.id,
            sort_by="TIMECREATED",
            sort_order="ASC",
        )
        # Can't delete the detault
        if len(securities) > 1:
            for security in securities[1:]:
                deleted = delete(
                    network_client,
                    "delete_security_list",
                    security.id,
                    wait_for_states=[SecurityList.LIFECYCLE_STATE_TERMINATED],
                )
                remove_stack["security_lists"].append(deleted)

        # Delete all DHCP options
        dhcp_options = list_entities(
            network_client,
            "list_dhcp_options",
            compartment_id,
            vcn.id,
            sort_by="TIMECREATED",
            sort_order="ASC",
        )

        if len(dhcp_options) > 1:
            for dhcp_option in dhcp_options[1:]:
                deleted = delete(
                    network_client,
                    "delete_dhcp_options",
                    dhcp_option.id,
                    wait_for_states=[DhcpOptions.LIFECYCLE_STATE_TERMINATED],
                )
                remove_stack["dhcp_options"].append(deleted)

        # Delete local peering gateways
        local_peering_gateways = list_entities(
            network_client, "list_local_peering_gateways", compartment_id, vcn.id
        )
        for local_peering_gateway in local_peering_gateways:
            deleted = delete(
                network_client,
                "delete_local_peering_gateway",
                local_peering_gateway.id,
                wait_for_states=[LocalPeeringGateway.LIFECYCLE_STATE_TERMINATED],
            )
            remove_stack["local_peering_gateways"].append(deleted)

        # Delete all NAT gateways
        nat_gateways = list_entities(
            network_client, "list_nat_gateways", compartment_id, vcn_id=vcn.id
        )
        for gateway in nat_gateways:
            deleted = delete(
                network_client,
                "delete_nat_gateway",
                gateway.id,
                wait_for_states=[NatGateway.LIFECYCLE_STATE_TERMINATED],
            )
            remove_stack["nat_gateways"].append(deleted)

        # Delete Service Gateways
        service_gateways = list_entities(
            network_client, "list_service_gateways", compartment_id, vcn_id=vcn.id
        )
        for service_gateway in service_gateways:
            deleted = delete(
                network_client,
                "delete_service_gateway",
                service_gateway.id,
                wait_for_states=[ServiceGateway.LIFECYCLE_STATE_TERMINATED],
            )
            remove_stack["service_gateways"].append(deleted)

        # The delete_vcn defaults internally to succeed_on_not_found
        # https://github.com/oracle/oci-python-sdk/blob/bafa4f0d68be097568772cd3cda250e60cb61a0c/src/oci/core/virtual_network_client_composite_operations.py#L1758
        deleted = delete(
            network_client,
            "delete_vcn",
            vcn.id,
            wait_for_states=[Vcn.LIFECYCLE_STATE_TERMINATED],
        )
        remove_stack["id"] = vcn.id
        remove_stack["vcn"] = deleted

    return remove_stack


def delete_compartment_vcns(network_client, compartment_id, **kwargs):
    removed_vcns = []
    vcns = list_entities(network_client, "list_vcns", compartment_id, **kwargs)
    for vcn in vcns:
        removed_stack = delete_vcn_stack(network_client, compartment_id, vcn_id=vcn.id)
        deleted = stack_was_deleted(removed_stack)
        removed_vcns.append(deleted)
    return removed_vcns


def get_subnet_by_name(network_client, compartment_id, vcn_id, display_name, **kwargs):
    subnets = list_entities(
        network_client, "list_subnets", compartment_id, vcn_id, **kwargs
    )
    for subnet in subnets:
        if subnet.display_name == display_name:
            return subnet


if __name__ == "__main__":
    vcn_args = get_arguments([VCN], strip_group_prefix=True)
    subnet_args = get_arguments([SUBNET], strip_group_prefix=True)
    args = get_arguments([OCI], strip_group_prefix=True)
    network_client = new_client(
        VirtualNetworkClient,
        composite_class=VirtualNetworkClientCompositeOperations,
        profile_name=args.profile_name,
    )

    removed_vcns = delete_compartment_vcns(network_client, args.compartment_id)
    print("Deleted vcns: {}".format(removed_vcns))



from corc.cli.parser import ParseKwargs


def valid_s3_group(parser):
    add_s3_group(parser)
    select_s3_group(parser)
    s3_config_group(parser)


def s3_config_group(parser):
    s3_config = parser.add_argument_group(title="S3 Config arguments")
    s3_config.add_argument("--s3-config-file", default="~/.aws/config")
    s3_config.add_argument("--s3-credentials-file", default="~/.aws/credentials")
    s3_config.add_argument("--s3-profile-name", default="default")


def select_s3_group(parser):
    s3_group = parser.add_argument_group(title="S3 Selection arguments")
    s3 = s3_group.add_mutually_exclusive_group(required=True)
    s3.add_argument("--s3-bucket-id", default="")
    s3.add_argument("--s3-bucket-name", default="")


def add_s3_group(parser):
    s3 = parser.add_argument_group(title="S3 arguments")
    s3.add_argument("--s3-bucket-input-prefix", default="input")
    s3.add_argument("--s3-bucket-output-prefix", default="output")


def list_s3_group(parser):
    s3_list = parser.add_argument_group(title="S3 List arguments")
    s3_list.add_argument("--s3-all", default=True, type=bool)


def s3_extra(parser):
    s3_extra = parser.add_argument_group(title="S3 Extra arguments")
    s3_extra.add_argument("--s3-extra-kwargs", nargs="*", action=ParseKwargs)



import argparse
import json
from corc.defaults import PACKAGE_NAME
from corc.cli.helpers import (
    add_cluster_cli,
    add_instance_cli,
    add_platform_group,
    job_cli,
)


def run():
    parser = argparse.ArgumentParser(prog=PACKAGE_NAME)
    # Platform
    add_platform_group(parser)

    commands = parser.add_subparsers(title="COMMAND")
    instance_parser = commands.add_parser("instance")
    add_instance_cli(instance_parser)

    cluster_parser = commands.add_parser("cluster")
    add_cluster_cli(cluster_parser)

    job_parser = commands.add_parser("job")
    job_cli(job_parser)

    args = parser.parse_args()
    # Execute default function
    if hasattr(args, "func"):
        result = args.func(args)
        if result:
            try:
                json_object = json.loads(result)
                print(json.dumps(json_object, indent=2))
            except Exception:
                print(result)
    return None


if __name__ == "__main__":
    arguments = run()



from corc.cli.job.result import results_group


def valid_job_group(parser):
    results_group(parser)
    job_meta_group(parser)
    job_group(parser)


def select_job_group(parser, required=True):
    job_group = parser.add_argument_group(title="Job Identity arguments")
    job_group.add_argument("--job-name", default="", required=required)


def job_meta_group(parser):
    job_group = parser.add_argument_group(title="Job Meta arguments")
    job_group.add_argument("--job-debug", action="store_true", default=False)
    job_group.add_argument("--job-env-override", action="store_true", default=True)
    job_group.add_argument("--job-num-jobs", default=1, type=int)
    job_group.add_argument("--job-num-parallel", default=1, type=int)


def job_group(parser):
    job_group = parser.add_argument_group(title="Job Execute arguments")
    job_group.add_argument("job_command", default="")
    job_group.add_argument("--job-name", default="")
    job_group.add_argument("--job-args", nargs="*", default="")
    job_group.add_argument("--job-capture", action="store_true", default=True)
    job_group.add_argument("--job-output-path", default="/tmp/output")



import copy
import oci
from oci.core.virtual_network_client import VirtualNetworkClient
from oci.core.virtual_network_client_composite_operations import (
    VirtualNetworkClientCompositeOperations,
)
from oci.container_engine import (
    ContainerEngineClient,
    ContainerEngineClientCompositeOperations,
)
from oci.container_engine.models import (
    CreateClusterDetails,
    CreateNodePoolDetails,
    NodePoolPlacementConfigDetails,
)
from cluster import (
    new_cluster_stack,
    get_cluster_by_name,
    delete_cluster_stack,
)
from helpers import new_client, get, get_kubernetes_version
from network import new_vcn_stack, get_vcn_stack, get_vcn_by_name
from args import get_arguments, OCI, CLUSTER, VCN, SUBNET, NODE


if __name__ == "__main__":
    oci_args = get_arguments([OCI], strip_group_prefix=True)
    vcn_args = get_arguments([VCN], strip_group_prefix=True)
    subnet_args = get_arguments([SUBNET], strip_group_prefix=True)
    cluster_args = get_arguments([CLUSTER], strip_group_prefix=True)
    node_args = get_arguments([NODE], strip_group_prefix=True)
    container_engine_client = new_client(
        ContainerEngineClient,
        composite_class=ContainerEngineClientCompositeOperations,
        profile_name=oci_args.profile_name,
    )

    network_client = new_client(
        VirtualNetworkClient,
        composite_class=VirtualNetworkClientCompositeOperations,
        profile_name=oci_args.profile_name,
    )

    stack = None
    # Find the target vcn (id or name)
    if vcn_args.id:
        stack = get_vcn_stack(network_client, oci_args.compartment_id, vcn_args.vcn_id)
    elif vcn_args.display_name:
        vcn = get_vcn_by_name(
            network_client, oci_args.compartment_id, vcn_args.display_name
        )
        if vcn:
            stack = get_vcn_stack(network_client, oci_args.compartment_id, vcn.id)
        else:
            vcn_kwargs = copy.deepcopy(vars(vcn_args))
            subnet_kwargs = copy.deepcopy(vars(subnet_args))
            vcn_kwargs.pop("id")
            subnet_kwargs.pop("id")

            stack = new_vcn_stack(
                network_client,
                oci_args.compartment_id,
                vcn_kwargs=vcn_kwargs,
                subnet_kwargs=subnet_kwargs,
            )
    else:
        exit(1)

    if not stack:
        exit(2)

    vcn = stack["vcn"]
    subnet = stack["subnets"][0]

    if not vcn:
        print("Failed to find a vcn")
        exit(1)

    if not cluster_args.kubernetes_version:
        kubernetes_version = get_kubernetes_version(container_engine_client)
    else:
        kubernetes_version = cluster_args.kubernetes_version

    cluster = get_cluster_by_name(
        container_engine_client, oci_args.compartment_id, cluster_args.name
    )

    if not cluster:
        # Create new stack
        create_cluster_details = CreateClusterDetails(
            compartment_id=oci_args.compartment_id,
            kubernetes_version=kubernetes_version,
            name=cluster_args.name,
            vcn_id=vcn.id,
        )

        node_pool_placement_config = NodePoolPlacementConfigDetails(
            availability_domain=node_args.availability_domain, subnet_id=subnet.id
        )

        node_config_detail = oci.container_engine.models.CreateNodePoolNodeConfigDetails(
            size=node_args.size, placement_configs=[node_pool_placement_config]
        )

        # Associate
        create_node_pool_details = CreateNodePoolDetails(
            compartment_id=oci_args.compartment_id,
            kubernetes_version=kubernetes_version,
            name=node_args.name,
            node_shape=node_args.shape,
            node_image_name=node_args.image_name,
            node_config_details=node_config_detail,
        )

        cluster_stack = new_cluster_stack(
            container_engine_client, create_cluster_details, create_node_pool_details
        )
        cluster = cluster_stack["cluster"]
    else:
        # Delete the cluster
        deleted = delete_cluster_stack(
            container_engine_client, cluster.id, delete_vcn=True
        )

        print(deleted)

    cluster = get(container_engine_client, "get_cluster", cluster.id)
    print(cluster)



class Scheduler:
    def provision_storage(self, config):
        raise NotImplementedError

    def prepare(self, config):
        raise NotImplementedError

    def submit(self, task):
        raise NotImplementedError

    def retrieve(self, job_id):
        raise NotImplementedError



# TODO, rename to `stage_endpoint` and `staging_path`
required_staging_fields = {
    "enable": bool,
    "endpoint": str,
    "credentials_path": str,
    "upload_path": str,
    "input_path": str,
    "output_path": str,
}

required_staging_values = {
    "enable": True,
    "endpoint": True,
    "credentials_path": True,
    "upload_path": False,
    "input_path": False,
    "output_path": True,
}


required_get_storage_fields = {
    "endpoint": str,
    "download_path": str,
}

required_get_storage_values = {"endpoint": True, "download_path": False}

required_delete_storage_fields = {
    "endpoint": str,
}

required_delete_storage_values = {"endpoint": True}



from kubernetes import client
from corc.kubernetes.config import load_kube_config

ROUND_ROBIN = "ROUND_ROBIN"


def list_nodes(api_instance, **kwargs):
    # Can accept label selector
    return api_instance.list_node(**kwargs)


class NodeManager:
    def __init__(self):
        loaded = load_kube_config()
        if not loaded:
            raise RuntimeError("Failed to load the kubernetes config")
        self.client = client.CoreV1Api()

        self.nodes = []
        self.last_selected_index = 0

    def discover(self, **kwargs):
        self.nodes = list_nodes(self.client, **kwargs).items

    def _select_round_robin(self):
        if not self.nodes:
            return None

        if (self.last_selected_index + 1) < len(self.nodes):
            next_index = self.last_selected_index + 1
            selected = self.nodes[next_index]
            self.last_selected_index = next_index
            return selected
        else:
            # Back to the start
            next_index = 0
            selected = self.nodes[next_index]
            self.last_selected_index = next_index
            return selected
        return None

    def select(self, selection_type=ROUND_ROBIN):
        if not self.nodes:
            return None

        if selection_type == ROUND_ROBIN:
            return self._select_round_robin()
        return None



import platform
import socket
import subprocess
import yaml


def parse_yaml(data):
    try:
        parsed = yaml.safe_load(data)
        return parsed
    except yaml.reader.ReaderError as err:
        print("Failed to parse yaml: {}".format(err))
    return False


def dump_yaml(path, data):
    try:
        with open(path, "w") as fh:
            yaml.dump(data, fh)
        return True
    except IOError as err:
        print("Failed to dump yaml: {} - {}".format(path, err))
    return False


def present_in(var, collection, verbose=False):
    if var not in collection:
        if verbose:
            print("variable: {} not present in: {}".format(var, collection))
        return False
    return True


def validate_dict_fields(input_dict, valid_fields, verbose=False, throw=False):

    for key, value in input_dict.items():
        if not present_in(key, valid_fields, verbose=verbose):
            if throw:
                raise KeyError(
                    "The key: {} is not allowed in: {}".format(key, valid_fields)
                )
            return False
        if not isinstance(value, (valid_fields[key],)):
            if verbose:
                print(
                    "variable: {}, value: {} is of incorrect type: {},"
                    " should be: {}".format(key, value, type(value), valid_fields[key])
                )
            if throw:
                raise TypeError(
                    "{}: should be {} but was: {}".format(
                        value, valid_fields[key], type(value)
                    )
                )
            return False
    return True


def validate_dict_types(input_dict, required_fields=None, verbose=False, throw=False):
    if not required_fields:
        required_fields = {}
    for var, _type in required_fields.items():
        if not present_in(var, input_dict, verbose=verbose):
            if throw:
                raise KeyError(
                    "Missing required: {} in {}".format(var, required_fields)
                )
            return False
        if not isinstance(input_dict[var], _type):
            if verbose:
                print(
                    "variable: {} value: {} is of incorrect type: {}".format(
                        var, input_dict[var], type(input_dict[var])
                    )
                )
            if throw:
                raise TypeError("{}: should be {}".format(var, _type))
            return False
    return True


def validate_dict_values(input_dict, required_values=None, verbose=False, throw=False):
    if not required_values:
        required_values = {}

    for var, required_value in required_values.items():
        if not present_in(var, input_dict, verbose=verbose):
            if throw:
                raise KeyError("Missing required: {} in {} ".format(var, input_dict))
            return False
        if required_value and not input_dict[var]:
            if verbose:
                print(
                    "The required variable: {} was not set in: {}".format(
                        var, input_dict
                    )
                )
            if throw:
                raise ValueError(
                    "{} doesn't have a value: {}".format(var, required_value)
                )
            return False
    return True


def ping(host):
    """
    # https://stackoverflow.com/questions/2953462/pinging-servers-in-python
    Returns True if host (str) responds to a ping request.
    Remember that a host may not respond to a ping (ICMP)
    request even if the host name is valid.
    """

    # Option for the number of packets as a function of
    param = "-n" if platform.system().lower() == "windows" else "-c"

    # Building the command. Ex: "ping -c 1 google.com"
    command = ["ping", param, "1", host]

    return subprocess.call(command) == 0


def open_port(host="127.0.0.1", port=22):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((host, port))
    sock.close()
    if result == 0:
        return True
    else:
        return False



def valid_subnet_group(parser):
    add_subnet_group(parser)


def add_subnet_group(parser):
    subnet_group = parser.add_argument_group(title="Subnet arguments")
    subnet_group.add_argument("--subnet-id", default="")
    subnet_group.add_argument("--subnet-dns-label", default=None)
    subnet_group.add_argument("--subnet-cidr-block", default="10.0.1.0/24")



def valid_vcn_group(parser):
    add_vcn_group(parser)


def add_vcn_group(parser):
    vcn_group = parser.add_argument_group(title="VCN arguments")
    vcn_group.add_argument("--vcn-id", default="")
    vcn_group.add_argument("--vcn-dns-label", default=None)
    vcn_group.add_argument("--vcn-display-name", default="VCN Network")
    vcn_group.add_argument("--vcn-cidr-block", default="10.0.0.0/16")



from corc.defaults import (
    OCI,
    SUBNET,
    COMPUTE,
    VCN,
)
from corc.cli.args import extract_arguments
from corc.oci.instance import launch_instance as oci_launch_instance


def launch_instance(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    compute_args = vars(extract_arguments(args, [COMPUTE]))
    subnet_args = vars(extract_arguments(args, [SUBNET]))
    vcn_args = vars(extract_arguments(args, [VCN]))

    if oci_args:
        return oci_launch_instance(
            compute_kwargs=compute_args,
            oci_kwargs=oci_args,
            subnet_kwargs=subnet_args,
            vcn_kwargs=vcn_args,
        )



default_cluster_config = {"name": "cluster", "kubernetes_version": "", "domain": ""}

default_configurer_config = {}

default_instance_config = {
    "ssh_authorized_keys": [],
    "availability_domain": "",
    "operating_system": "CentOS",
    "operating_system_version": "7",
    "target_shape": "VM.Standard2.1",
}

default_job_config = {
    "meta": {
        "name": "",
        "debug": False,
        "env_override": True,
        "num_jobs": 1,
        "num_parallel": 1,
    },
    "capture": True,
    "output_path": "/tmp/output",
}

default_network_config = {
    "subnet": {"id": "", "dns_label": None, "cidr_block": "10.0.1.0/24"},
    "vcn": {
        "id": "",
        "dns_label": None,
        "display_name": "VCN Network",
        "cidr_block": "10.0.0.0/16",
    },
}

default_platform_config = {"oci": {"profile_name": "DEFAULT", "compartment_id": ""}}

default_storage_config = {
    "s3": {
        "config_file": "~/.aws/config",
        "credentials_file": "~/.aws/credentials",
        "profile_name": "default",
        "bucket_id": "",
        "bucket_name": "",
        "bucket_input_prefix": "input",
        "bucket_output_prefix": "output",
    },
    "endpoint": "",
    "credentials_path": "/mnt/creds",
    "upload_path": "",
    "input_path": "/tmp/input",
    "output_path": "/tmp/output",
    "download_path": "",
}


default_config = {
    "cluster": default_cluster_config,
    "configurer": default_configurer_config,
    "instance": default_instance_config,
    "job": default_job_config,
    "network": default_network_config,
    "platform": default_platform_config,
    "storage": default_storage_config,
}

valid_cluster_config = {
    "image": str,
    "id": str,
    "name": str,
    "kubernetes_version": str,
    "domain": str,
    "node": dict,
}

valid_node_config = {
    "id": str,
    "name": str,
    "availability_domain": str,
    "size": int,
    "shape": str,
    "image": str,
}

valid_configurer_config = {
    "root_path": str,
    "playbook_path": str,
    "inventory_path": str,
}

valid_instance_config = {
    "ssh_authorized_keys": list,
    "availability_domain": str,
    "operating_system": str,
    "operating_system_version": str,
    "target_shape": str,
}


valid_job_config = {
    "meta": {
        "name": str,
        "debug": bool,
        "env_override": bool,
        "num_jobs": int,
        "num_parallel": int,
    },
    "command": str,
    "args": list,
    "capture": bool,
    "output_path": str,
}


valid_network_config = {
    "subnet": {"id": str, "dns_label": str, "cidr_block": str},
    "vcn": {"id": str, "dns_label": str, "display_name": str, "cidr_block": str},
}

valid_platform_config = {"profile_name": str, "compartment_id": str}

valid_s3_config = {
    "config_file": str,
    "credentials_file": str,
    "profile_name": str,
    "bucket_id": str,
    "bucket_name": str,
    "bucket_input_prefix": str,
    "bucket_output_prefix": str,
}

valid_storage_config = {
    "s3": dict,
    "endpoint": str,
    "credentials_path": str,
    "upload_path": str,
    "input_path": str,
    "output_path": str,
    "download_path": str,
}


valid_corc_config = {
    "cluster": valid_cluster_config,
    "configurer": valid_configurer_config,
    "instance": valid_instance_config,
    "job": valid_job_config,
    "network": valid_network_config,
    "platform": valid_platform_config,
    "storage": valid_storage_config,
}



class Orchestrator:

    options = None
    _is_ready = False
    _is_reachable = False

    def __init__(self, options):
        self.options = options

    def is_ready(self):
        return self._is_ready

    def is_reachable(self):
        return self._is_reachable

    def endpoint(self, select=None):
        raise NotImplementedError

    def poll(self):
        raise NotImplementedError

    def setup(self):
        raise NotImplementedError

    def tear_down(self):
        raise NotImplementedError

    @classmethod
    def validate_options(cls, options):
        raise NotImplementedError



import os
import yaml
from corc.util import validate_dict_fields
from corc.config.defaults import valid_corc_config, default_config


def generate_default_config():
    return default_config


def save_config(config, path=None):
    if "CORC_CONFIG_FILE" in os.environ:
        path = os.environ["CORC_CONFIG_FILE"]
    else:
        if not path:
            # Ensure the directory path is there
            path = os.path.join(os.path.expanduser("~"), ".corc", "config")
            dir_path = os.path.dirname(path)
            if not os.path.exists(dir_path):
                try:
                    os.makedirs(os.path.dirname(path))
                except Exception as err:
                    print("Failed to create config directory: {}".format(err))

    if not config:
        return False

    try:
        with open(path, "w") as fh:
            yaml.safe_dump(config, fh)
    except Exception as err:
        print("Failed to save config: {}".format(err))
        return False
    return True


def load_config(path=None):
    config = {}
    if "CORC_CONFIG_FILE" in os.environ:
        path = os.environ["CORC_CONFIG_FILE"]
    else:
        if not path:
            path = os.path.join(os.path.expanduser("~"), ".corc", "config")

    if not os.path.exists(path):
        return False
    try:
        with open(path, "r") as fh:
            config = yaml.safe_load(fh)
    except Exception as err:
        print("Failed to load config: {}".format(err))
    return config


def remove_config(path=None):
    if "CORC_CONFIG_FILE" in os.environ:
        path = os.environ["CORC_CONFIG_FILE"]
    else:
        if not path:
            path = os.path.join(os.path.expanduser("~"), ".corc", "config")

    if not os.path.exists(path):
        return True
    try:
        os.remove(path)
    except Exception as err:
        print("Failed to remove config: {}".format(err))
        return False
    return True


def valid_config(config, verbose=False, throw=False):
    if not isinstance(config, dict):
        return False

    for key, value in config.items():
        if key not in valid_corc_config:
            return False
        valid = validate_dict_fields(
            value, valid_corc_config[key], verbose=verbose, throw=throw
        )

        if not valid:
            return False

    return True



import os
import unittest
from oci.core import (
    ComputeClient,
    ComputeClientCompositeOperations,
    VirtualNetworkClient,
    VirtualNetworkClientCompositeOperations,
)
from oci.container_engine import (
    ContainerEngineClient,
    ContainerEngineClientCompositeOperations,
)
from corc.oci.helpers import new_client, get, get_kubernetes_version
from corc.oci.cluster import (
    new_cluster_stack,
    get_cluster_stack,
    get_cluster_by_name,
    delete_cluster_stack,
    valid_cluster_stack,
    gen_cluster_stack_details,
    list_entities,
)
from corc.oci.network import new_vcn_stack, delete_vcn_stack, valid_vcn_stack


class TestClusterStack(unittest.TestCase):
    def setUp(self):
        # Load compartment_id from the env
        if "OCI_COMPARTMENT_ID" not in os.environ:
            raise ValueError("Missing required environment variable OCI_COMPARTMENT_ID")

        if "OCI_PROFILE_NAME" in os.environ:
            profile_name = os.environ["OCI_PROFILE_NAME"]
        else:
            profile_name = "DEFAULT"

        oci_options = dict(
            compartment_id=os.environ["OCI_COMPARTMENT_ID"], profile_name=profile_name,
        )

        test_name = "Test_Cluster"
        cluster_name = test_name
        node_name = test_name + "_Node"
        vcn_name = test_name + "_Network"
        subnet_name = test_name + "_Subnet"

        # Add unique test postfix
        if "OCI_TEST_ID" in os.environ:
            cluster_name += os.environ["OCI_TEST_ID"]
            node_name += os.environ["OCI_TEST_ID"]
            vcn_name += os.environ["OCI_TEST_ID"]
            subnet_name += os.environ["OCI_TEST_ID"]

        image_options = dict(display_name="Oracle-Linux-7.7-2020.03.23-0",)
        node_options = dict(
            availability_domain="lfcb:EU-FRANKFURT-1-AD-1",
            name=node_name,
            size=1,
            node_shape="VM.Standard1.1",
            image=image_options,
        )
        vcn_options = dict(
            cidr_block="10.0.0.0/16", display_name=vcn_name, dns_label="ku",
        )

        subnet_options = dict(display_name=subnet_name, dns_label="workers")

        self.container_engine_client = new_client(
            ContainerEngineClient,
            composite_class=ContainerEngineClientCompositeOperations,
            profile_name=profile_name,
        )

        cluster_options = dict(
            name=cluster_name,
            kubernetes_version=get_kubernetes_version(self.container_engine_client),
        )

        self.compute_client = new_client(
            ComputeClient,
            composite_class=ComputeClientCompositeOperations,
            profile_name=profile_name,
        )

        self.network_client = new_client(
            VirtualNetworkClient,
            composite_class=VirtualNetworkClientCompositeOperations,
            profile_name=profile_name,
        )

        self.options = dict(
            oci=oci_options,
            cluster=cluster_options,
            node=node_options,
            vcn=vcn_options,
            subnet=subnet_options,
        )

    def tearDown(self):
        # Validate that the vcn stack is gone
        deleted = delete_cluster_stack(
            self.container_engine_client, self.cluster_stack["id"], delete_vcn=True
        )
        self.assertTrue(deleted)

        # Delete all vcn with display_name
        vcns = list_entities(
            self.network_client,
            "list_vcns",
            self.options["oci"]["compartment_id"],
            display_name=self.options["vcn"]["display_name"],
        )

        for vcn in vcns:
            deleted = delete_vcn_stack(
                self.network_client, self.options["oci"]["compartment_id"], vcn=vcn
            )
            self.assertTrue(deleted)

    def test_cluster_stack(self):
        # Need vcn stack for the cluster stack
        self.vcn_stack = new_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )
        self.assertTrue(valid_vcn_stack(self.vcn_stack))

        # Available images
        available_images = list_entities(
            self.compute_client,
            "list_images",
            self.options["oci"]["compartment_id"],
            **self.options["node"]["image"]
        )

        if not available_images:
            raise ValueError(
                "No valid image could be found with options: {}".format(
                    self.options["image"]
                )
            )

        if len(available_images) > 1:
            raise ValueError(
                "More than 1 image was found with options: {}".format(
                    self.options["image"]
                )
            )

        image = available_images[0]

        # Prepare cluster details
        cluster_details = gen_cluster_stack_details(
            self.vcn_stack["id"], self.vcn_stack["subnets"], image, **self.options
        )

        self.cluster_stack = new_cluster_stack(
            self.container_engine_client,
            cluster_details["create_cluster"],
            cluster_details["create_node_pool"],
        )

        self.assertTrue(valid_cluster_stack(self.cluster_stack))

        _cluster_stack = get_cluster_stack(
            self.container_engine_client,
            self.options["oci"]["compartment_id"],
            self.cluster_stack["id"],
        )
        self.assertTrue(valid_cluster_stack(_cluster_stack))
        self.assertEqual(self.cluster_stack["id"], _cluster_stack["id"])
        self.assertEqual(self.cluster_stack["cluster"], _cluster_stack["cluster"])
        self.assertListEqual(
            self.cluster_stack["node_pools"], _cluster_stack["node_pools"]
        )

        new_cluster_id = get(
            self.container_engine_client, "get_cluster", self.cluster_stack["id"]
        )

        new_cluster_name = get_cluster_by_name(
            self.container_engine_client,
            self.options["oci"]["compartment_id"],
            self.cluster_stack["cluster"].name,
        )

        self.assertEqual(self.cluster_stack["id"], new_cluster_id.id)
        self.assertEqual(self.cluster_stack["cluster"], new_cluster_name)


if __name__ == "__main__":
    unittest.main()



import os
import time
import unittest
from corc.configurer import AnsibleConfigurer
from corc.oci.instance import OCIInstanceOrchestrator


current_dir = os.path.dirname(os.path.abspath(__file__))
playbook_path = os.path.join(current_dir, "res", "configurer", "playbook.yml")


class TestInstanceConfigurer(unittest.TestCase):
    def setUp(self):
        # Load compartment_id from the env
        if "OCI_COMPARTMENT_ID" not in os.environ:
            raise ValueError("Missing required environment variable OCI_COMPARTMENT_ID")

        if "OCI_PROFILE_NAME" in os.environ:
            profile_name = os.environ["OCI_PROFILE_NAME"]
        else:
            profile_name = "DEFAULT"

        oci_options = dict(
            compartment_id=os.environ["OCI_COMPARTMENT_ID"], profile_name=profile_name,
        )

        test_name = "Test_Instance_Conf"
        node_name = test_name + "_Node"
        vcn_name = test_name + "_Network"
        subnet_name = test_name + "_Subnet"

        # Add unique test postfix
        if "OCI_TEST_ID" in os.environ:
            node_name += os.environ["OCI_TEST_ID"]
            vcn_name += os.environ["OCI_TEST_ID"]
            subnet_name += os.environ["OCI_TEST_ID"]

        compute_options = dict(
            availability_domain="lfcb:EU-FRANKFURT-1-AD-1",
            shape="VM.Standard1.1",
            operating_system="CentOS",
            operating_system_version="7",
            display_name=node_name,
        )

        self.ssh_private_key_file = None
        ssh_public_key_file = None
        ssh_public_key = None

        if "OCI_INSTANCE_SSH_KEY" in os.environ:
            self.ssh_private_key_file = os.environ["OCI_INSTANCE_SSH_KEY"]
            if not os.path.exists(self.ssh_private_key_file):
                raise ValueError(
                    "The specified OCI_INSTANCE_SSH_KEY path: {} does not exist".format(
                        self.ssh_private_key_file
                    )
                )

            # Get the public key compliment
            ssh_public_key_file = self.ssh_private_key_file + ".pub"
            with open(ssh_public_key_file, "r") as pub_file:
                ssh_public_key = pub_file.read()

        if not ssh_public_key_file or not self.ssh_private_key_file:
            raise ValueError(
                "Failed to load ssh keys, OCI_INSTANCE_SSH_KEY "
                "environment variable must be set"
            )

        if not ssh_public_key:
            raise ValueError(
                "Failed to load the specified OCI_INSTANCE_SSH_KEY public key"
            )

        compute_metadata_options = dict(ssh_authorized_keys=[ssh_public_key])

        vcn_options = dict(
            cidr_block="10.0.0.0/16", display_name=vcn_name, dns_label="ku",
        )
        subnet_options = dict(display_name=subnet_name, dns_label="workers")

        self.options = dict(
            oci=oci_options,
            compute=compute_options,
            compute_metadata=compute_metadata_options,
            vcn=vcn_options,
            subnet=subnet_options,
        )

        OCIInstanceOrchestrator.validate_options(self.options)
        self.orchestrator = OCIInstanceOrchestrator(self.options)
        # Should not be ready at this point
        self.orchestrator.setup()
        self.assertTrue(self.orchestrator.is_ready())
        # Ensure that the orchestrated instance is reachable
        reachable = self.orchestrator.is_reachable()
        num_waited, max_wait = 0, 60

        while not reachable:
            self.orchestrator.poll()
            reachable = self.orchestrator.is_reachable()
            if num_waited > max_wait:
                print(
                    "The maximum number of waits: '{}' for the orchestrated instance to "
                    "be reachable was exceeded".format(max_wait)
                )
                break
            time.sleep(1)
            num_waited += 1
        self.assertTrue(reachable)

    def tearDown(self):
        self.orchestrator.tear_down()
        self.assertFalse(self.orchestrator.is_ready())
        self.orchestrator = None
        self.options = None

    def test_instance_ansible_configure(self):
        # Extract the ip of the instance
        endpoint = self.orchestrator.endpoint()
        options = dict(
            ssh_private_key_file=self.ssh_private_key_file,
            playbook_path=playbook_path,
            hosts=[endpoint],
        )
        configurer = AnsibleConfigurer(options)
        configurer.apply()


if __name__ == "__main__":
    unittest.main()



from oci.core import (
    ComputeClient,
    ComputeClientCompositeOperations,
    VirtualNetworkClient,
    VirtualNetworkClientCompositeOperations,
)
from oci.container_engine import (
    ContainerEngineClient,
    ContainerEngineClientCompositeOperations,
)
from oci.container_engine.models import Cluster, CreateClusterDetails, WorkRequest
from oci.container_engine.models import (
    CreateNodePoolDetails,
    NodePoolPlacementConfigDetails,
    CreateNodePoolNodeConfigDetails,
    NodeSourceViaImageDetails,
)

from corc.oci.helpers import (
    new_client,
    create,
    delete,
    get,
    list_entities,
    get_kubernetes_version,
)
from corc.util import parse_yaml
from corc.kubernetes.config import save_kube_config, load_kube_config
from corc.orchestrator import Orchestrator
from corc.oci.network import (
    new_vcn_stack,
    get_vcn_stack,
    valid_vcn_stack,
    get_vcn_by_name,
    delete_vcn_stack,
    refresh_vcn_stack,
)


def new_cluster_engine_client(**kwargs):
    return new_client(
        ContainerEngineClient,
        composite_class=ContainerEngineClientCompositeOperations,
        **kwargs,
    )


def refresh_kube_config(cluster_id, profile_name="DEFAULT"):
    container_engine_client = new_client(
        ContainerEngineClient,
        composite_class=ContainerEngineClientCompositeOperations,
        profile_name=profile_name,
    )

    # Try to load the existing
    # Create or refresh the kubernetes config
    kube_config = create(container_engine_client, "create_kubeconfig", cluster_id)
    if kube_config and hasattr(kube_config, "text"):
        config_dict = parse_yaml(kube_config.text)
        # HACK, add profile to user args
        if profile_name:
            profile_args = ["--profile", profile_name]
            config_dict["users"][0]["user"]["exec"]["args"].extend(profile_args)
        if save_kube_config(config_dict):
            loaded = load_kube_config()

    loaded = load_kube_config()
    if not loaded:
        # The new/refreshed config failed to load
        return False
    return True


# FIXME, no node_pools should be allowed
def valid_cluster_stack(stack):

    if not isinstance(stack, dict):
        raise TypeError("The Cluster stack must be a dictionary to be valid")

    expected_fields = ["id", "cluster", "node_pools"]
    for field in expected_fields:
        if field not in stack:
            return False
        # Value can't be None/False
        if not stack[field]:
            return False
    return True


def new_cluster_stack(
    container_engine_client,
    create_cluster_details,
    create_node_pool_details,
    discover_vcn=False,
):
    stack = dict(id=None, cluster=None, node_pools=[])
    cluster = create_cluster(container_engine_client, create_cluster_details)

    if not cluster:
        return stack

    # cluster
    stack["cluster"] = cluster
    if hasattr(cluster, "id"):
        stack["id"] = cluster.id
        create_node_pool_details.cluster_id = cluster.id

    node_pool = create_node_pool(container_engine_client, create_node_pool_details)

    if node_pool:
        stack["node_pools"].append(node_pool)

    return stack


def update_cluster_stack(
    container_engine_client, update_cluster_details, update_node_pool_details
):
    raise NotImplementedError


def get_cluster_stack(
    container_engine_client, compartment_id, cluster_id, node_kwargs=None
):

    if not node_kwargs:
        node_kwargs = dict()

    stack = dict(id=None, cluster=None, node_pools=[])

    cluster = get(container_engine_client, "get_cluster", cluster_id)
    if not cluster:
        return stack

    stack["cluster"] = cluster
    if hasattr(cluster, "id"):
        stack["id"] = cluster.id

    # Get node pools (NodePoolSummaries)
    node_pool_summaries = list_entities(
        container_engine_client,
        "list_node_pools",
        compartment_id,
        cluster_id=cluster.id,
        **node_kwargs,
    )

    for summary in node_pool_summaries:
        node_pool = get(container_engine_client, "get_node_pool", summary.id)
        if node_pool:
            stack["node_pools"].append(node_pool)
    return stack


def delete_cluster_stack(container_engine_client, cluster_id, delete_vcn=False):
    # Will implicitly delete the node pool as well
    cluster = get(container_engine_client, "get_cluster", cluster_id)
    if not cluster:
        return False
    return delete_cluster(container_engine_client, cluster_id)


def delete_cluster(container_engine_client, cluster_id, **kwargs):
    return delete(
        container_engine_client,
        "delete_cluster",
        cluster_id,
        wait_for_states=[WorkRequest.STATUS_SUCCEEDED, WorkRequest.STATUS_FAILED],
        **kwargs,
    )


def create_cluster(container_engine_client, create_cluster_details, create_kwargs=None):
    if not create_kwargs:
        create_kwargs = dict(
            wait_for_states=[
                WorkRequest.STATUS_SUCCEEDED,
                WorkRequest.OPERATION_TYPE_CLUSTER_CREATE,
                WorkRequest.STATUS_FAILED,
            ]
        )

    created_response = create(
        container_engine_client,
        "create_cluster",
        create_cluster_details=create_cluster_details,
        **create_kwargs,
    )

    if not created_response:
        return None

    cluster = None
    # NOTE, only supports a single cluster creation for now
    if created_response.resources:
        cluster_id = created_response.resources[0].identifier
        # Get the actual created cluster
        cluster = get(container_engine_client, "get_cluster", cluster_id)

    return cluster


def list_clusters(container_engine_client, compartment_id, cluster_kwargs=None):

    if not cluster_kwargs:
        cluster_kwargs = {}

    if "lifecycle_state" not in cluster_kwargs:
        cluster_kwargs.update(dict(lifecycle_state=[Cluster.LIFECYCLE_STATE_ACTIVE]))

    # ClusterSummaries
    return list_entities(
        container_engine_client,
        "list_clusters",
        compartment_id=compartment_id,
        **cluster_kwargs,
    )


def get_cluster_by_name(
    container_engine_client, compartment_id, name, cluster_kwargs=None
):
    if not cluster_kwargs:
        cluster_kwargs = {}

    cluster_kwargs.update(dict(name=name))

    clusters = list_clusters(container_engine_client, compartment_id, cluster_kwargs)
    if clusters:
        # Convert to cluster type
        cluster = get(container_engine_client, "get_cluster", clusters[0].id)
        return cluster
    return None


def delete_node_pool(container_engine_client, cluster_id, **kwargs):
    return delete(container_engine_client, "delete_node_pool", cluster_id, **kwargs)


def create_node_pool(container_engine_client, create_node_pool_details):
    created_response = create(
        container_engine_client,
        "create_node_pool",
        wait_for_states=[
            WorkRequest.STATUS_SUCCEEDED,
            WorkRequest.OPERATION_TYPE_NODEPOOL_CREATE,
            WorkRequest.STATUS_FAILED,
        ],
        create_node_pool_details=create_node_pool_details,
    )
    if not created_response:
        return None

    node_pool = None
    # NOTE, only supports a single cluster creation for now
    if created_response.resources:
        node_pool_id = created_response.resources[0].identifier
        # Get the actual created cluster
        node_pool = get(container_engine_client, "get_node_pool", node_pool_id)

    return node_pool


def prepare_node_source_details(image, **kwargs):
    if not image:
        raise RuntimeError("No image was provided")

    node_source_details = {
        k: v for k, v in kwargs.items() if hasattr(NodeSourceViaImageDetails, k)
    }

    node_source_details = NodeSourceViaImageDetails(
        image_id=image.id, **node_source_details
    )

    return node_source_details


def prepare_create_cluster_details(**kwargs):
    cluster_kwargs = {
        k: v for k, v in kwargs.items() if hasattr(CreateClusterDetails, k)
    }
    create_cluster_details = CreateClusterDetails(**cluster_kwargs)
    return create_cluster_details


def prepare_node_pool_placement_config(**kwargs):
    node_pool_kwargs = {
        k: v for k, v in kwargs.items() if hasattr(NodePoolPlacementConfigDetails, k)
    }

    return NodePoolPlacementConfigDetails(**node_pool_kwargs)


def prepare_create_node_pool_config(**kwargs):
    create_node_pool_config_kwargs = {
        k: v for k, v in kwargs.items() if hasattr(CreateNodePoolNodeConfigDetails, k)
    }
    return CreateNodePoolNodeConfigDetails(**create_node_pool_config_kwargs)


def prepare_create_node_pool_details(**kwargs):
    # Discover the image_id if the name is provided
    create_node_pool_details = {
        k: v for k, v in kwargs.items() if hasattr(CreateNodePoolDetails, k)
    }
    return CreateNodePoolDetails(**create_node_pool_details)


def gen_cluster_stack_details(vnc_id, subnets, image, **options):
    cluster_details = {}

    create_cluster_details = prepare_create_cluster_details(
        vcn_id=vnc_id, **options["oci"], **options["cluster"],
    )
    cluster_details["create_cluster"] = create_cluster_details

    node_pool_place_configs = []
    if subnets:
        for subnet in subnets:
            node_pool_place_configs.append(
                prepare_node_pool_placement_config(
                    subnet_id=subnet.id, **options["node"]
                )
            )

    node_config_details = prepare_create_node_pool_config(
        size=options["node"]["size"], placement_configs=node_pool_place_configs,
    )

    # Convert the deprecrated image name, to the proper source details
    node_source_details = prepare_node_source_details(image)
    if node_source_details:
        options["node"]["node_source_details"] = node_source_details

    create_node_pool_details = prepare_create_node_pool_details(
        node_config_details=node_config_details,
        kubernetes_version=options["cluster"]["kubernetes_version"],
        **options["oci"],
        **options["node"],
    )
    cluster_details["create_node_pool"] = create_node_pool_details

    return cluster_details


class OCIClusterOrchestrator(Orchestrator):
    def __init__(self, options):
        super().__init__(options)
        # Set clients
        self.compute_client = new_client(
            ComputeClient,
            composite_class=ComputeClientCompositeOperations,
            profile_name=options["oci"]["profile_name"],
        )
        self.container_engine_client = new_client(
            ContainerEngineClient,
            composite_class=ContainerEngineClientCompositeOperations,
            profile_name=options["oci"]["profile_name"],
        )
        self.network_client = new_client(
            VirtualNetworkClient,
            composite_class=VirtualNetworkClientCompositeOperations,
            profile_name=options["oci"]["profile_name"],
        )

        if (
            "kubernetes_version" not in self.options["cluster"]
            or not self.options["cluster"]["kubernetes_version"]
        ):
            self.options["cluster"]["kubernetes_version"] = get_kubernetes_version(
                self.container_engine_client
            )

        self.cluster_stack = None
        self.vcn_stack = None
        self._is_ready = False

    def _get_vcn_stack(self):
        if "id" in self.options["vcn"] and self.options["vcn"]["id"]:
            return get_vcn_stack(
                self.network_client,
                self.options["oci"]["compartment_id"],
                self.options["vcn"]["id"],
            )

        stack = {}
        if (
            "display_name" in self.options["vcn"]
            and self.options["vcn"]["display_name"]
        ):
            vcn = get_vcn_by_name(
                self.network_client,
                self.options["oci"]["compartment_id"],
                self.options["vcn"]["display_name"],
            )
            if vcn:
                stack = get_vcn_stack(
                    self.network_client, self.options["oci"]["compartment_id"], vcn.id
                )
        return stack

    def _new_vcn_stack(self):
        stack = new_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )
        return stack

    def _refresh_vcn_stack(self, vcn_stack):
        stack = refresh_vcn_stack(
            self.network_client,
            self.options["oci"]["compartment_id"],
            vcn_kwargs=self.options["vcn"],
            subnet_kwargs=self.options["subnet"],
        )
        return stack

    def poll(self):
        pass

    def setup(self):
        # Ensure we have a VCN stack ready
        vcn_stack = self._get_vcn_stack()
        if not vcn_stack:
            vcn_stack = self._new_vcn_stack()

        if valid_vcn_stack(vcn_stack):
            self.vcn_stack = vcn_stack
        else:
            self.vcn_stack = self._refresh_vcn_stack(vcn_stack)

        if not valid_vcn_stack(self.vcn_stack):
            raise RuntimeError(
                "Failed to fix the broken VCN stack: {}".format(vcn_stack)
            )

        # Available images
        available_images = list_entities(
            self.compute_client,
            "list_images",
            self.options["oci"]["compartment_id"],
            **self.options["node"]["image"],
        )

        if not available_images:
            raise ValueError(
                "No valid image could be found with options: {}".format(
                    self.options["image"]
                )
            )

        if len(available_images) > 1:
            raise ValueError(
                "More than 1 image was found with options: {}".format(
                    self.options["image"]
                )
            )

        image = available_images[0]

        cluster_details = gen_cluster_stack_details(
            self.vcn_stack["id"], self.vcn_stack["subnets"], image, **self.options,
        )

        cluster = get_cluster_by_name(
            self.container_engine_client,
            self.options["oci"]["compartment_id"],
            self.options["cluster"]["name"],
        )
        if not cluster:
            self._is_ready = False
            # Ensure that we don't change the state options
            cluster_stack = new_cluster_stack(
                self.container_engine_client,
                cluster_details["create_cluster"],
                cluster_details["create_node_pool"],
            )
            if valid_cluster_stack(cluster_stack):
                self.cluster_stack = cluster_stack
            else:
                raise ValueError(
                    "The new cluster stack: {} is not valid".format(cluster_stack)
                )
        else:
            cluster_stack = get_cluster_stack(
                self.container_engine_client,
                self.options["oci"]["compartment_id"],
                cluster.id,
            )

            if cluster_stack["cluster"] and not cluster_stack["node_pools"]:
                cluster_details["create_node_pool"].cluster_id = cluster_stack[
                    "cluster"
                ].id
                node_pool = create_node_pool(
                    self.container_engine_client, cluster_details["create_node_pool"]
                )
                if node_pool:
                    cluster_stack["node_pools"].append(node_pool)

            if valid_cluster_stack(cluster_stack):
                self.cluster_stack = cluster_stack

        if self.cluster_stack:
            self._is_ready = True

    def tear_down(self):
        if not self.cluster_stack:
            # refresh
            self.cluster_stack = get_cluster_by_name(
                self.container_engine_client,
                self.options["oci"]["compartment_id"],
                self.options["cluster"]["name"],
            )

        if self.cluster_stack:
            cluster = self.cluster_stack["cluster"]
            deleted = delete_cluster_stack(self.container_engine_client, cluster.id)
            if deleted:
                self.cluster_stack = None
        else:
            self.cluster_stack = None

        if not self.vcn_stack:
            # refresh
            self.vcn_stack = self._get_vcn_stack()

        if self.vcn_stack:
            vcn_deleted = delete_vcn_stack(
                self.network_client,
                self.options["oci"]["compartment_id"],
                vcn_id=self.vcn_stack["id"],
            )
            if vcn_deleted:
                self.vcn_stack = None

        if self.cluster_stack and self.vcn_stack:
            self._is_ready = True
        else:
            self._is_ready = False

    @classmethod
    def validate_options(cls, options):
        if not isinstance(options, dict):
            raise TypeError("options is not a dictionary")

        expected_oci_keys = [
            "compartment_id",
            "profile_name",
        ]

        expected_cluster_keys = [
            "name",
        ]
        optional_cluster_keys = ["kubernetes_version", "domain"]

        expected_node_keys = [
            "availability_domain",
            "name",
            "size",
            "node_shape",
            "image",
        ]
        optional_node_keys = ["ssh_public_key"]

        expected_vcn_keys = ["cidr_block"]
        optional_vcn_keys = ["id", "display_name", "dns_label"]

        expected_subnet_keys = ["cidr_block"]
        optional_subnet_keys = ["id", "dns_label", "display_name"]

        expected_groups = {
            "oci": expected_oci_keys,
            "cluster": expected_cluster_keys + optional_cluster_keys,
            "node": expected_node_keys + optional_node_keys,
            "vcn": expected_vcn_keys + optional_vcn_keys,
            "subnet": expected_subnet_keys + optional_subnet_keys,
        }

        for group, keys in expected_groups.items():
            if group not in options:
                raise KeyError("Missing group: {}".format(group))

            if not isinstance(options[group], dict):
                raise TypeError("Group: {} must be a dictionary".format(group))

            for key, _ in options[group].items():
                if key not in keys:
                    raise KeyError("Incorrect key: {} is not in: {}".format(key, keys))



def valid_oci_group(parser):
    add_oci_group(parser)


def add_oci_group(parser):
    oci_group = parser.add_argument_group(title="OCI arguments")
    oci_group.add_argument("--oci-profile-name", default="DEFAULT")
    oci_group.add_argument("--oci-compartment-id", default="")



from corc.defaults import OCI, CLUSTER, SUBNET, VCN, NODE
from corc.cli.args import extract_arguments
from corc.oci.cluster import (
    list_clusters as oci_list_clusters,
    delete_cluster_stack as oci_delete_cluster_stack,
    get_cluster_by_name as oci_get_cluster_by_name,
)
from corc.oci.cluster import (
    OCIClusterOrchestrator,
    new_cluster_engine_client,
)


def list_clusters(args):
    oci_args = vars(extract_arguments(args, [OCI]))

    if oci_args:
        container_engine_client = new_cluster_engine_client(
            profile_name=oci_args["profile_name"]
        )
        return oci_list_clusters(container_engine_client, oci_args["compartment_id"])


def start_cluster(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    cluster_args = vars(extract_arguments(args, [CLUSTER]))
    node_args = vars(extract_arguments(args, [NODE]))
    vcn_args = vars(extract_arguments(args, [VCN]))
    subnet_args = vars(extract_arguments(args, [SUBNET]))

    if oci_args:
        # Validate oci authentication
        # Node shape is special
        if "shape" in node_args:
            node_args["node_shape"] = node_args["shape"]
            node_args.pop("shape")

        image_options = {}
        if "image" in node_args:
            image_name = node_args["image"]
            image_options = dict(display_name=image_name)
            node_args.pop("image")

        node_args["image"] = image_options

        cluster_options = dict(
            oci=oci_args,
            cluster=cluster_args,
            node=node_args,
            vcn=vcn_args,
            subnet=subnet_args,
        )
        OCIClusterOrchestrator.validate_options(cluster_options)
        orchestrator = OCIClusterOrchestrator(cluster_options)
        orchestrator.setup()


def stop_cluster(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    cluster_args = vars(extract_arguments(args, [CLUSTER]))

    if oci_args:
        # Discover the vcn_stack for the cluster
        if not cluster_args["id"] and not cluster_args["name"]:
            raise ValueError("Either the id or name of the cluster must" "be provided")

        container_engine_client = new_cluster_engine_client(
            profile_name=oci_args["profile_name"]
        )
        if cluster_args["id"]:
            cluster_id = cluster_args["id"]
        else:
            cluster = oci_get_cluster_by_name(
                container_engine_client,
                oci_args["compartment_id"],
                cluster_args["name"],
            )
            cluster_id = cluster.id

        return oci_delete_cluster_stack(container_engine_client, cluster_id)


def update_cluster(args):
    oci_args = vars(extract_arguments(args, [OCI]))
    # cluster_args = vars(extract_arguments(args, [CLUSTER]))
    # node_args = vars(extract_arguments(args, [NODGE]))
    # vcn_args = vars(extract_arguments(args, [VCN]))
    # subnet_args = vars(extract_arguments(args, [SUBNET]))

    if oci_args:
        pass


def get_cluster(args):
    raise NotImplementedError



import boto3
import os
import time
from oci.container_engine import (
    ContainerEngineClient,
    ContainerEngineClientCompositeOperations,
)
from kubernetes import client
from kubernetes.client import (
    V1Volume,
    V1SecretVolumeSource,
    V1VolumeMount,
    V1ObjectMeta,
)
from kubernetes.client.rest import ApiException

from jobio.storage.s3 import expand_s3_bucket, stage_s3_resource
from corc.defaults import (
    STORAGE_CREDENTIALS_NAME,
    KUBERNETES_NAMESPACE,
    JOB_DEFAULT_NAME,
)
from corc.util import validate_dict_types, validate_dict_values
from corc.storage.staging import (
    required_staging_fields,
    required_staging_values,
    required_get_storage_fields,
    required_get_storage_values,
    required_delete_storage_fields,
    required_delete_storage_values,
)
from corc.storage.s3 import (
    bucket_exists,
    delete_bucket,
    delete_objects,
    list_objects,
    load_s3_config,
    required_s3_fields,
    required_s3_values,
    upload_to_s3,
    upload_directory_to_s3,
)
from corc.kubernetes.nodes import NodeManager
from corc.kubernetes.scheduler import KubenetesScheduler
from corc.oci.cluster import get_cluster_by_name, refresh_kube_config
from corc.oci.helpers import new_client


def run(
    cluster_kwargs={},
    execute_kwargs={},
    job_kwargs={},
    oci_kwargs={},
    staging_kwargs={},
    storage_kwargs={},
):
    if "name" not in job_kwargs or not job_kwargs["name"]:
        since_epoch = int(time.time())
        job_kwargs["name"] = "{}-{}".format(JOB_DEFAULT_NAME, since_epoch)

    container_engine_client = new_client(
        ContainerEngineClient,
        composite_class=ContainerEngineClientCompositeOperations,
        profile_name=oci_kwargs["profile_name"],
    )

    cluster = get_cluster_by_name(
        container_engine_client,
        oci_kwargs["compartment_id"],
        name=cluster_kwargs["name"],
    )

    if not cluster:
        print("Failed to find a cluster with name: {}".format(cluster_kwargs["name"]))
        return False

    refreshed = refresh_kube_config(cluster.id, profile_name=oci_kwargs["profile_name"])
    if not refreshed:
        exit(1)

    node_manager = NodeManager()
    node_manager.discover()
    node = node_manager.select()

    # Ensure we have the newest config
    scheduler = KubenetesScheduler()

    jobio_args = [
        "jobio",
        "run",
        execute_kwargs["command"],
        "--execute-args",
        " ".join(execute_kwargs["args"]),
        "--execute-output-path",
        execute_kwargs["output_path"],
        "--job-name",
        job_kwargs["name"],
    ]

    if "capture" in execute_kwargs and execute_kwargs["capture"]:
        jobio_args.append("--execute-capture")

    if "debug" in execute_kwargs:
        jobio_args.append("--job-debug")

    if "env_override" in job_kwargs:
        jobio_args.append("--job-env-override")

    # Maintained by the pod
    volumes = []
    # Maintained by the container
    volume_mounts = []
    # Environment to pass to the container
    envs = []

    # Prepare config for the scheduler
    scheduler_config = {}

    if staging_kwargs and staging_kwargs["enable"]:
        validate_dict_types(staging_kwargs, required_staging_fields, throw=True)
        validate_dict_values(staging_kwargs, required_staging_values, throw=True)

        # Means that results should be exported to the specified storage
        # Create kubernetes secrets
        core_api = client.CoreV1Api()
        # storage_api = client.StorageV1Api()

        # Storage endpoint credentials secret
        try:
            storage_credentials_secret = core_api.read_namespaced_secret(
                STORAGE_CREDENTIALS_NAME, KUBERNETES_NAMESPACE
            )
        except ApiException as err:
            storage_credentials_secret = None
            print(
                "Failed to find the secret: {}, err: {}".format(
                    STORAGE_CREDENTIALS_NAME, err
                )
            )

        # volumes
        secret_volume_source = V1SecretVolumeSource(
            secret_name=STORAGE_CREDENTIALS_NAME
        )
        secret_volume = V1Volume(
            name=STORAGE_CREDENTIALS_NAME, secret=secret_volume_source
        )
        volumes.append(secret_volume)

        # Where the storage credentials should be mounted
        # in the compute unit
        secret_mount = V1VolumeMount(
            name=STORAGE_CREDENTIALS_NAME,
            mount_path=staging_kwargs["credentials_path"],
            read_only=True,
        )
        volume_mounts.append(secret_mount)

        if storage_kwargs:
            validate_dict_types(storage_kwargs, required_s3_fields, throw=True)
            validate_dict_values(storage_kwargs, required_s3_values, throw=True)
            # S3 storage
            # Look for s3 credentials and config files
            s3_config = load_s3_config(
                storage_kwargs["config_file"],
                storage_kwargs["credentials_file"],
                staging_kwargs["endpoint"],
                profile_name=storage_kwargs["profile_name"],
            )

            if not storage_credentials_secret:
                secret_data = dict(
                    aws_access_key_id=s3_config["aws_access_key_id"],
                    aws_secret_access_key=s3_config["aws_secret_access_key"],
                )
                secret_metadata = V1ObjectMeta(name=STORAGE_CREDENTIALS_NAME)
                secrets_config = dict(metadata=secret_metadata, string_data=secret_data)
                scheduler_config.update(dict(secret_kwargs=secrets_config))

            # If `access_key`
            # TODO, unify argument endpoint, with s3 config endpoint'
            s3 = boto3.resource("s3", **s3_config)

            bucket = bucket_exists(s3.meta.client, storage_kwargs["bucket_name"])
            if not bucket:
                bucket = s3.create_bucket(
                    Bucket=storage_kwargs["bucket_name"],
                    CreateBucketConfiguration={
                        "LocationConstraint": s3_config["region_name"]
                    },
                )

            if "upload_path" in staging_kwargs and staging_kwargs["upload_path"]:
                # Upload local path to the bucket as designated input for the job
                if os.path.exists(staging_kwargs["upload_path"]):
                    if os.path.isdir(staging_kwargs["upload_path"]):
                        uploaded = upload_directory_to_s3(
                            s3.meta.client,
                            staging_kwargs["upload_path"],
                            storage_kwargs["bucket_name"],
                            s3_prefix=storage_kwargs["bucket_input_prefix"],
                        )
                    elif os.path.isfile(staging_kwargs["upload_path"]):
                        s3_path = os.path.basename(staging_kwargs["upload_path"])
                        if storage_kwargs["bucket_input_prefix"]:
                            s3_path = os.path.join(
                                storage_kwargs["bucket_input_prefix"], s3_path
                            )
                        # Upload
                        uploaded = upload_to_s3(
                            s3.meta.client,
                            staging_kwargs["upload_path"],
                            s3_path,
                            storage_kwargs["bucket_name"],
                        )

                if not uploaded:
                    raise RuntimeError(
                        "Failed to local path: {} in the upload folder to s3".format(
                            staging_kwargs["upload_path"]
                        )
                    )

            jobio_args.extend(
                [
                    "--s3-region-name",
                    s3_config["region_name"],
                    "--storage-secrets-dir",
                    staging_kwargs["credentials_path"],
                    "--storage-endpoint",
                    staging_kwargs["endpoint"],
                    "--storage-input-path",
                    staging_kwargs["input_path"],
                    "--storage-output-path",
                    staging_kwargs["output_path"],
                    "--bucket-name",
                    storage_kwargs["bucket_name"],
                    "--bucket-input-prefix",
                    storage_kwargs["bucket_input_prefix"],
                    "--bucket-output-prefix",
                    storage_kwargs["bucket_output_prefix"],
                ]
            )

            if "enable" in staging_kwargs:
                jobio_args.append("--storage-enable")

            # Provide a way to allow pod specific output prefixes
            # field_ref = client.V1ObjectFieldSelector(field_path="metadata.name")
            field_ref = client.V1ObjectFieldSelector(field_path="metadata.name")
            env_var_source = client.V1EnvVarSource(field_ref=field_ref)
            # HACK, Set the output prefix in the bucket to the name of the pod
            env_output_prefix = client.V1EnvVar(
                name="JOBIO_BUCKET_OUTPUT_PREFIX", value_from=env_var_source
            )
            envs.append(env_output_prefix)

    if scheduler_config:
        prepared = scheduler.prepare(**scheduler_config)
        if not prepared:
            raise RuntimeError("Failed to prepare the scheduler")

    container_spec = dict(
        name=job_kwargs["name"],
        image=cluster_kwargs["image"],
        env=envs,
        args=jobio_args,
        volume_mounts=volume_mounts,
    )
    # args=jobio_args,
    pod_spec = dict(node_name=node.metadata.name, volumes=volumes, dns_policy="Default")

    job_spec = dict(
        backoff_limit=2,
        parallelism=job_kwargs["num_parallel"],
        completions=job_kwargs["num_jobs"],
    )

    task = dict(
        container_kwargs=container_spec,
        pod_spec_kwargs=pod_spec,
        job_spec_kwargs=job_spec,
    )

    job_id = scheduler.submit(**task)
    if not job_id:
        exit(1)


def get_results(job_kwargs={}, staging_kwargs={}, storage_kwargs={}):

    validate_dict_types(staging_kwargs, required_get_storage_fields, throw=True)
    validate_dict_values(staging_kwargs, required_get_storage_values, throw=True)

    validate_dict_types(storage_kwargs, required_s3_fields, throw=True)
    validate_dict_values(storage_kwargs, required_s3_values, throw=True)

    # S3 storage
    # Look for s3 credentials and config files
    s3_config = load_s3_config(
        storage_kwargs["config_file"],
        storage_kwargs["credentials_file"],
        staging_kwargs["endpoint"],
        profile_name=storage_kwargs["profile_name"],
    )

    # Download results from s3
    s3_resource = stage_s3_resource(**s3_config)

    # Whether to expand all or a single result
    if "result_prefix" in job_kwargs and job_kwargs["result_prefix"]:
        result_prefix = job_kwargs["result_prefix"]
    else:
        # Use all
        result_prefix = ""

    bucket = bucket_exists(s3_resource.meta.client, job_kwargs["name"])
    if not bucket:
        raise RuntimeError(
            "Could not find a bucket with the name: {}".format(job_kwargs["name"])
        )

    expanded = expand_s3_bucket(
        s3_resource,
        job_kwargs["name"],
        staging_kwargs["download_path"],
        s3_prefix=result_prefix,
    )

    if not expanded:
        raise RuntimeError(
            "Failed to expand the target bucket: {}".format(job_kwargs["name"])
        )


def delete_results(job_kwargs={}, staging_kwargs={}, storage_kwargs={}):

    validate_dict_types(staging_kwargs, required_delete_storage_fields, throw=True)
    validate_dict_values(staging_kwargs, required_delete_storage_values, throw=True)

    validate_dict_types(storage_kwargs, required_s3_fields, throw=True)
    validate_dict_values(storage_kwargs, required_s3_values, throw=True)

    # S3 storage
    # Look for s3 credentials and config files
    s3_config = load_s3_config(
        storage_kwargs["config_file"],
        storage_kwargs["credentials_file"],
        staging_kwargs["endpoint"],
        profile_name=storage_kwargs["profile_name"],
    )

    # Download results from s3
    s3_resource = stage_s3_resource(**s3_config)

    # Whether to expand all or a single result
    if "result_prefix" in job_kwargs and job_kwargs["result_prefix"]:
        result_prefix = job_kwargs["result_prefix"]
    else:
        # Use all
        result_prefix = ""

    bucket = bucket_exists(s3_resource.meta.client, job_kwargs["name"])
    if not bucket:
        raise RuntimeError(
            "Could not find a bucket with the name: {}".format(job_kwargs["name"])
        )

    deleted = delete_objects(s3_resource, job_kwargs["name"], s3_prefix=result_prefix)

    if "Errors" in deleted:
        for error in deleted["Errors"]:
            print("Failed to delete: {}".format(error))

    # If the bucket is empty, remove it as well
    results = list_objects(s3_resource, job_kwargs["name"])

    if not results:
        if not delete_bucket(s3_resource.meta.client, job_kwargs["name"]):
            return False
    return True


def list_results(
    job_kwargs={}, staging_kwargs={}, storage_kwargs={}, storage_extra_kwargs={},
):
    # S3 storage
    # Look for s3 credentials and config files
    s3_config = load_s3_config(
        storage_kwargs["config_file"],
        storage_kwargs["credentials_file"],
        staging_kwargs["endpoint"],
        profile_name=storage_kwargs["profile_name"],
    )

    s3_resource = stage_s3_resource(**s3_config)
    response = {}
    results = []
    if "name" in job_kwargs and job_kwargs["name"]:
        bucket = bucket_exists(s3_resource.meta.client, job_kwargs["name"])
        if not bucket:
            raise RuntimeError(
                "Could not find a bucket with the name: {}".format(job_kwargs["name"])
            )
        results = list_objects(s3_resource, job_kwargs["name"], **storage_extra_kwargs)
    else:
        response = s3_resource.meta.client.list_buckets()
        if "Buckets" in response:
            results = response["Buckets"]
    return results



from kubernetes import client


def prepare_job(container_kwargs=None, pod_spec_kwargs=None, job_spec_kwargs=None):

    if not container_kwargs:
        container_kwargs = {}

    if not pod_spec_kwargs:
        pod_spec_kwargs = {}

    if not job_spec_kwargs:
        job_spec_kwargs = {}

    if "name" not in container_kwargs:
        print("missing name from container_kwargs")
        return False

    container = client.V1Container(**container_kwargs)

    # Round robin
    # V1PodSpec Can use node_selector to target specific node
    # Create and configurate a spec section
    template = client.V1PodTemplateSpec(
        metadata=client.V1ObjectMeta(generate_name=container_kwargs["name"]),
        spec=client.V1PodSpec(
            restart_policy="Never", containers=[container], **pod_spec_kwargs
        ),
    )
    # Create the specification of deployment
    spec = client.V1JobSpec(template=template, **job_spec_kwargs)
    # Instantiate the job object
    job = client.V1Job(
        api_version="batch/v1",
        kind="Job",
        metadata=client.V1ObjectMeta(generate_name=container_kwargs["name"]),
        spec=spec,
    )
    return job


def create_job(api_instance, job):
    api_response = api_instance.create_namespaced_job(body=job, namespace="default")
    return api_response


# def update_job(api_instance, job):
#     # Update container image
#     job.spec.template.spec.containers[0].image = "perl"
#     api_response = api_instance.patch_namespaced_job(
#         name=JOB_NAME, namespace="default", body=job
#     )
#     print("Job updated. status='%s'" % str(api_response.status))


# def delete_job(api_instance):
#     api_response = api_instance.delete_namespaced_job(
#         name=JOB_NAME,
#         namespace="default",
#         body=client.V1DeleteOptions(
#             propagation_policy="Foreground", grace_period_seconds=5
#         ),
#     )
#     print("Job deleted. status='%s'" % str(api_response.status))
